Package: RMariaDB
Title: Database Interface and MariaDB Driver
Version: 1.3.5
Authors@R: c(
    person("Kirill", "Müller", , "kirill@cynkra.com", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0002-1416-3412")),
    person("Jeroen", "Ooms", role = "aut",
           comment = c(ORCID = "0000-0002-4035-0289")),
    person("David", "James", role = "aut"),
    person("Saikat", "DebRoy", role = "aut"),
    person("Hadley", "Wickham", role = "aut"),
    person("Jeffrey", "Horner", role = "aut"),
    person("R Consortium", role = "fnd"),
    person("RStudio", role = "cph")
  )
Description: Implements a DBI-compliant interface to MariaDB
    (<https://mariadb.org/>) and MySQL (<https://www.mysql.com/>)
    databases.
License: MIT + file LICENSE
URL: https://rmariadb.r-dbi.org, https://github.com/r-dbi/RMariaDB,
        https://downloads.mariadb.org/connector-c/
BugReports: https://github.com/r-dbi/RMariaDB/issues
Depends: R (>= 2.8.0)
Imports: bit64, blob, DBI (>= 1.1.3), hms (>= 0.5.0), lubridate,
        methods, rlang
Suggests: DBItest (>= 1.7.2.9001), decor, readr, rprojroot, testthat
        (>= 3.0.0), withr
LinkingTo: cpp11
Config/Needs/website: r-dbi/dbitemplate
Config/autostyle/scope: line_breaks
Config/autostyle/strict: false
Config/testthat/edition: 3
Encoding: UTF-8
NeedsCompilation: yes
RoxygenNote: 7.3.3.9000
SystemRequirements: libmariadb-client-lgpl-dev or libmariadb-dev or
        libmysqlclient-dev, with libssl-dev (deb),
        mariadb-connector-c-devel or mariadb-devel (rpm),
        mariadb-connector-c or mysql-connector-c (brew)
Collate: 'MariaDBConnection.R' 'MariaDBDriver.R' 'MariaDBResult.R'
        'RMariaDB-package.R' 'coerce.R' 'compatRowNames.R' 'connect.R'
        'cpp11.R' 'dbAppendTable_MariaDBConnection.R'
        'dbBegin_MariaDBConnection.R' 'dbBind_MariaDBResult.R'
        'dbClearResult_MariaDBResult.R' 'dbColumnInfo_MariaDBResult.R'
        'dbCommit_MariaDBConnection.R' 'dbConnect_MariaDBDriver.R'
        'dbDataType_MariaDBConnection.R' 'dbDataType_MariaDBDriver.R'
        'dbDisconnect_MariaDBConnection.R'
        'dbExistsTable_MariaDBConnection_character.R'
        'dbFetch_MariaDBResult.R' 'dbGetInfo_MariaDBConnection.R'
        'dbGetInfo_MariaDBDriver.R' 'dbGetRowCount_MariaDBResult.R'
        'dbGetRowsAffected_MariaDBResult.R'
        'dbGetStatement_MariaDBResult.R'
        'dbHasCompleted_MariaDBResult.R'
        'dbIsValid_MariaDBConnection.R' 'dbIsValid_MariaDBDriver.R'
        'dbIsValid_MariaDBResult.R'
        'dbListObjects_MariaDBConnection_ANY.R'
        'dbListTables_MariaDBConnection.R'
        'dbQuoteIdentifier_MariaDBConnection_Id.R'
        'dbQuoteIdentifier_MariaDBConnection_SQL.R'
        'dbQuoteIdentifier_MariaDBConnection_character.R'
        'dbQuoteLiteral_MariaDBConnection.R'
        'dbQuoteLiteral_MySQLConnection.R'
        'dbQuoteString_MariaDBConnection_SQL.R'
        'dbQuoteString_MariaDBConnection_character.R'
        'dbReadTable_MariaDBConnection_character.R'
        'dbRemoveTable_MariaDBConnection_character.R'
        'dbRollback_MariaDBConnection.R'
        'dbSendQuery_MariaDBConnection_character.R'
        'dbSendStatement_MariaDBConnection_character.R'
        'dbUnloadDriver_MariaDBDriver.R'
        'dbUnquoteIdentifier_MariaDBConnection_SQL.R'
        'dbWriteTable_MariaDBConnection_character_character.R'
        'dbWriteTable_MariaDBConnection_character_data.frame.R'
        'default.R' 'export.R' 'query.R' 'quote-mysql.R' 'quote.R'
        'show_MariaDBConnection.R' 'sqlData_MariaDBConnection.R'
        'table.R' 'transaction.R' 'utils.R' 'zzz.R'
Packaged: 2026-02-05 19:48:36 UTC; kirill
Author: Kirill Müller [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-1416-3412>),
  Jeroen Ooms [aut] (ORCID: <https://orcid.org/0000-0002-4035-0289>),
  David James [aut],
  Saikat DebRoy [aut],
  Hadley Wickham [aut],
  Jeffrey Horner [aut],
  R Consortium [fnd],
  RStudio [cph]
Maintainer: Kirill Müller <kirill@cynkra.com>
Repository: CRAN
Date/Publication: 2026-02-06 06:40:54 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-04-27 21:41:58 UTC; unix
Archs: RMariaDB.so.dSYM

Package: RMariaDB
Title: Database Interface and MariaDB Driver
Version: 1.3.5
Authors@R: c(
    person("Kirill", "Müller", , "kirill@cynkra.com", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0002-1416-3412")),
    person("Jeroen", "Ooms", role = "aut",
           comment = c(ORCID = "0000-0002-4035-0289")),
    person("David", "James", role = "aut"),
    person("Saikat", "DebRoy", role = "aut"),
    person("Hadley", "Wickham", role = "aut"),
    person("Jeffrey", "Horner", role = "aut"),
    person("R Consortium", role = "fnd"),
    person("RStudio", role = "cph")
  )
Description: Implements a DBI-compliant interface to MariaDB
    (<https://mariadb.org/>) and MySQL (<https://www.mysql.com/>)
    databases.
License: MIT + file LICENSE
URL: https://rmariadb.r-dbi.org, https://github.com/r-dbi/RMariaDB,
        https://downloads.mariadb.org/connector-c/
BugReports: https://github.com/r-dbi/RMariaDB/issues
Depends: R (>= 2.8.0)
Imports: bit64, blob, DBI (>= 1.1.3), hms (>= 0.5.0), lubridate,
        methods, rlang
Suggests: DBItest (>= 1.7.2.9001), decor, readr, rprojroot, testthat
        (>= 3.0.0), withr
LinkingTo: cpp11
Config/Needs/website: r-dbi/dbitemplate
Config/autostyle/scope: line_breaks
Config/autostyle/strict: false
Config/testthat/edition: 3
Encoding: UTF-8
NeedsCompilation: yes
RoxygenNote: 7.3.3.9000
SystemRequirements: libmariadb-client-lgpl-dev or libmariadb-dev or
        libmysqlclient-dev, with libssl-dev (deb),
        mariadb-connector-c-devel or mariadb-devel (rpm),
        mariadb-connector-c or mysql-connector-c (brew)
Collate: 'MariaDBConnection.R' 'MariaDBDriver.R' 'MariaDBResult.R'
        'RMariaDB-package.R' 'coerce.R' 'compatRowNames.R' 'connect.R'
        'cpp11.R' 'dbAppendTable_MariaDBConnection.R'
        'dbBegin_MariaDBConnection.R' 'dbBind_MariaDBResult.R'
        'dbClearResult_MariaDBResult.R' 'dbColumnInfo_MariaDBResult.R'
        'dbCommit_MariaDBConnection.R' 'dbConnect_MariaDBDriver.R'
        'dbDataType_MariaDBConnection.R' 'dbDataType_MariaDBDriver.R'
        'dbDisconnect_MariaDBConnection.R'
        'dbExistsTable_MariaDBConnection_character.R'
        'dbFetch_MariaDBResult.R' 'dbGetInfo_MariaDBConnection.R'
        'dbGetInfo_MariaDBDriver.R' 'dbGetRowCount_MariaDBResult.R'
        'dbGetRowsAffected_MariaDBResult.R'
        'dbGetStatement_MariaDBResult.R'
        'dbHasCompleted_MariaDBResult.R'
        'dbIsValid_MariaDBConnection.R' 'dbIsValid_MariaDBDriver.R'
        'dbIsValid_MariaDBResult.R'
        'dbListObjects_MariaDBConnection_ANY.R'
        'dbListTables_MariaDBConnection.R'
        'dbQuoteIdentifier_MariaDBConnection_Id.R'
        'dbQuoteIdentifier_MariaDBConnection_SQL.R'
        'dbQuoteIdentifier_MariaDBConnection_character.R'
        'dbQuoteLiteral_MariaDBConnection.R'
        'dbQuoteLiteral_MySQLConnection.R'
        'dbQuoteString_MariaDBConnection_SQL.R'
        'dbQuoteString_MariaDBConnection_character.R'
        'dbReadTable_MariaDBConnection_character.R'
        'dbRemoveTable_MariaDBConnection_character.R'
        'dbRollback_MariaDBConnection.R'
        'dbSendQuery_MariaDBConnection_character.R'
        'dbSendStatement_MariaDBConnection_character.R'
        'dbUnloadDriver_MariaDBDriver.R'
        'dbUnquoteIdentifier_MariaDBConnection_SQL.R'
        'dbWriteTable_MariaDBConnection_character_character.R'
        'dbWriteTable_MariaDBConnection_character_data.frame.R'
        'default.R' 'export.R' 'query.R' 'quote-mysql.R' 'quote.R'
        'show_MariaDBConnection.R' 'sqlData_MariaDBConnection.R'
        'table.R' 'transaction.R' 'utils.R' 'zzz.R'
Packaged: 2026-02-05 19:48:36 UTC; kirill
Author: Kirill Müller [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-1416-3412>),
  Jeroen Ooms [aut] (ORCID: <https://orcid.org/0000-0002-4035-0289>),
  David James [aut],
  Saikat DebRoy [aut],
  Hadley Wickham [aut],
  Jeffrey Horner [aut],
  R Consortium [fnd],
  RStudio [cph]
Maintainer: Kirill Müller <kirill@cynkra.com>
Repository: CRAN
Date/Publication: 2026-02-06 06:40:54 UTC

# Generated by roxygen2: do not edit by hand

S3method(format,MariaDBConnection)
export(CLIENT_COMPRESS)
export(CLIENT_CONNECT_WITH_DB)
export(CLIENT_FOUND_ROWS)
export(CLIENT_IGNORE_SIGPIPE)
export(CLIENT_IGNORE_SPACE)
export(CLIENT_INTERACTIVE)
export(CLIENT_LOCAL_FILES)
export(CLIENT_LONG_FLAG)
export(CLIENT_LONG_PASSWORD)
export(CLIENT_MULTI_RESULTS)
export(CLIENT_MULTI_STATEMENTS)
export(CLIENT_NO_SCHEMA)
export(CLIENT_ODBC)
export(CLIENT_PROTOCOL_41)
export(CLIENT_RESERVED)
export(CLIENT_RESERVED2)
export(CLIENT_SSL)
export(CLIENT_SSL_VERIFY_SERVER_CERT)
export(CLIENT_TRANSACTIONS)
export(Id)
export(MariaDB)
export(mariadbClientLibraryVersions)
export(mariadbDefault)
export(mariadbHasDefault)
exportClasses(MariaDBConnection)
exportClasses(MariaDBDriver)
exportClasses(MariaDBResult)
exportClasses(MySQLConnection)
exportMethods(dbAppendTable)
exportMethods(dbBegin)
exportMethods(dbBind)
exportMethods(dbCanConnect)
exportMethods(dbClearResult)
exportMethods(dbColumnInfo)
exportMethods(dbCommit)
exportMethods(dbConnect)
exportMethods(dbCreateTable)
exportMethods(dbDataType)
exportMethods(dbDisconnect)
exportMethods(dbDriver)
exportMethods(dbExecute)
exportMethods(dbExistsTable)
exportMethods(dbFetch)
exportMethods(dbGetInfo)
exportMethods(dbGetQuery)
exportMethods(dbGetRowCount)
exportMethods(dbGetRowsAffected)
exportMethods(dbGetStatement)
exportMethods(dbHasCompleted)
exportMethods(dbIsReadOnly)
exportMethods(dbIsValid)
exportMethods(dbListFields)
exportMethods(dbListObjects)
exportMethods(dbListResults)
exportMethods(dbListTables)
exportMethods(dbQuoteIdentifier)
exportMethods(dbQuoteLiteral)
exportMethods(dbQuoteString)
exportMethods(dbReadTable)
exportMethods(dbRemoveTable)
exportMethods(dbRollback)
exportMethods(dbSendQuery)
exportMethods(dbSendStatement)
exportMethods(dbUnloadDriver)
exportMethods(dbUnquoteIdentifier)
exportMethods(dbWithTransaction)
exportMethods(dbWriteTable)
exportMethods(show)
import(DBI)
import(methods)
importFrom(DBI,Id)
importFrom(bit64,integer64)
importFrom(bit64,is.integer64)
importFrom(hms,hms)
importFrom(utils,read.table)
importFrom(utils,write.table)
useDynLib(RMariaDB, .registration = TRUE)

RMariaDB
======

<!-- badges: start -->
[![rcc](https://github.com/r-dbi/RMariaDB/workflows/rcc/badge.svg)](https://github.com/r-dbi/RMariaDB/actions)
[![Codecov test coverage](https://codecov.io/gh/r-dbi/RMariaDB/branch/main/graph/badge.svg)](https://app.codecov.io/gh/r-dbi/RMariaDB?branch=main)
[![CRAN_Status_Badge](https://www.r-pkg.org/badges/version/RMariaDB)](https://cran.r-project.org/package=RMariaDB)
[![CRAN RStudio mirror downloads](https://cranlogs.r-pkg.org/badges/RMariaDB)](https://cran.r-project.org/package=RMariaDB)
<!-- badges: end -->

RMariaDB is a database interface and MariaDB driver for R. This version is aimed at full compliance with the [DBI specification](https://cran.r-project.org/package=DBI/vignettes/spec.html), as a replacement for the old [RMySQL](https://cran.r-project.org/package=RMySQL) package.

## Hello World

```R
library(DBI)
# Connect to my-db as defined in ~/.my.cnf
con <- dbConnect(RMariaDB::MariaDB(), group = "my-db")

dbListTables(con)
dbWriteTable(con, "mtcars", mtcars)
dbListTables(con)

dbListFields(con, "mtcars")
dbReadTable(con, "mtcars")

# You can fetch all results:
res <- dbSendQuery(con, "SELECT * FROM mtcars WHERE cyl = 4")
dbFetch(res)
dbClearResult(res)

# Or a chunk at a time
res <- dbSendQuery(con, "SELECT * FROM mtcars WHERE cyl = 4")
while(!dbHasCompleted(res)){
  chunk <- dbFetch(res, n = 5)
  print(nrow(chunk))
}
# Clear the result
dbClearResult(res)

# Disconnect from the database
dbDisconnect(con)
```

## Installation

Binary packages for __OS X__ or __Windows__ can be installed directly from CRAN:

```r
install.packages("RMariaDB")
```

The development version from github:

```R
# install.packages("devtools")
devtools::install_github("r-dbi/DBI")
devtools::install_github("r-dbi/RMariaDB")
```

Discussions associated with DBI and related database packages take place on [R-SIG-DB](https://stat.ethz.ch/mailman/listinfo/r-sig-db).
The website [Databases using R](https://db.rstudio.com/) describes the tools and best practices in this ecosystem.

Installation from source on Linux or OS X currently requires [`MariaDB Connector/C`](https://downloads.mariadb.org/connector-c/), preferably in version 2.3.4/3.0.3 or later.
With older versions, character and blob columns do not work reliably.
Alternatively, Oracle's [libmysqlclient](https://packages.debian.org/buster/default-libmysqlclient-dev) can be used.


### Connector/C

On recent __Debian__ or __Ubuntu__ install [libmariadb-dev](https://packages.debian.org/testing/libmariadb-dev).

```
sudo apt-get install -y libmariadb-dev
```

On __Fedora__,  __CentOS__ or __RHEL__ we need [mariadb-devel](https://src.fedoraproject.org/rpms/mariadb):

```
sudo yum install mariadb-devel
````

On __OS X__ use [mariadb-connector-c](https://github.com/Homebrew/homebrew-core/blob/master/Formula/m/mariadb-connector-c.rb) from Homebrew:

```
brew install mariadb-connector-c
```


### MySQL client library

On recent __Debian__ or __Ubuntu__ install [libmysqlclient-dev](https://packages.debian.org/buster/default-libmysqlclient-dev).

```
sudo apt-get install -y libmysqlclient-dev
```

On __Fedora__,  __CentOS__ or __RHEL__ we need mysql-devel, see https://apps.fedoraproject.org/packages/mysql-devel:

```
sudo yum install mysql-devel
```

Follow [instructions](https://dev.mysql.com/doc/mysql-yum-repo-quick-guide/en/) to enable the MySQL yum repository if the above command attempts to install MariaDB files.


On __OS X__ use [mysql-connector-c](https://github.com/Homebrew/homebrew-core/blob/master/Formula/m/mysql-connector-c++.rb) from Homebrew:

```
brew install mysql-connector-c++
```


## MariaDB configuration file

Instead of specifying a username and password in calls to `dbConnect()`, it's better to set up a MariaDB configuration file that names the databases that you connect to most commonly. This file should live in `~/.my.cnf` and look like:

```
[database_name]
option1=value1
option2=value2
```

If you want to run the examples, you'll need to set the proper options in the `[rs-dbi]` group of any MariaDB option file, such as /etc/my.cnf or the .my.cnf file in your home directory. For a default single user install of MariaDB, the following code should work:

```
[rs-dbi]
database="test"
user="root"
password=""
```

## Acknowledgements

Many thanks to Christoph M. Friedrich, John Heuer, Kurt Hornik, Torsten Hothorn, Saikat Debroy, Matthew Kelly, Brian D. Ripley, Mikhail Kondrin, Jake Luciani, Jens Nieschulze, Deepayan Sarkar, Louis Springer, Duncan Temple Lang, Luis Torgo, Arend P. van der Veen, Felix Weninger, J. T. Lindgren, Crespin Miller, and Michal Okonlewski, Seth Falcon and Paul Gilbert for comments, suggestions, bug reports, and patches to the original [RMySQL](https://cran.r-project.org/package=RMySQL) package, and to all contributors (of [code](https://github.com/r-dbi/RMariaDB/graphs/contributors) and discussions) to this package.

---

Please note that the 'RMariaDB' project is released with a
[Contributor Code of Conduct](https://rmariadb.r-dbi.org/CODE_OF_CONDUCT.html).
By contributing to this project, you agree to abide by its terms.

#' @rdname mariadb-tables
#' @importFrom utils read.table
#' @param sep field separator character
#' @param eol End-of-line separator
#' @param skip number of lines to skip before reading data in the input file.
#' @param quote the quote character used in the input file (defaults to
#'    `\"`.)
#' @param header logical, does the input file have a header line? Default is the
#'    same heuristic used by `read.table()`, i.e., `TRUE` if the first
#'    line has one fewer column that the second line.
#' @param nrows number of lines to rows to import using `read.table` from
#'   the input file to create the proper table definition. Default is 50.
#' @usage NULL
dbWriteTable_MariaDBConnection_character_character <- function(conn, name, value, field.types = NULL, overwrite = FALSE,
                                                               append = FALSE, header = TRUE, row.names = FALSE, nrows = 50,
                                                               sep = ",", eol = "\n", skip = 0, quote = '"', temporary = FALSE,
                                                               ...) {
  if (overwrite && append)
    stop("overwrite and append cannot both be TRUE", call. = FALSE)

  found <- dbExistsTable(conn, name)
  if (found && !overwrite && !append) {
    stop("Table ", name, " exists in database, and both overwrite and",
      " append are FALSE",
      call. = FALSE
    )
  }
  if (found && overwrite) {
    dbRemoveTable(conn, name)
  }
  if (!found && append) {
    stop("Table ", name, " does not exists when appending")
  }

  if (!found || overwrite) {
    if (is.null(field.types)) {
      # Initialise table with first `nrows` lines
      d <- read.table(value,
        sep = sep, header = header, skip = skip,
        nrows = nrows, na.strings = "\\N", comment.char = "",
        stringsAsFactors = FALSE
      )
      field.types <- vapply(d, dbDataType,
        dbObj = conn,
        FUN.VALUE = character(1)
      )
    }

    sql <- sqlCreateTable(conn, name, field.types,
      row.names = row.names, temporary = temporary
    )
    dbExecute(conn, sql)
  }

  path <- normalizePath(value, winslash = "/", mustWork = TRUE)
  sql <- paste0(
    "LOAD DATA LOCAL INFILE ", dbQuoteString(conn, path), "\n",
    "INTO TABLE ", dbQuoteIdentifier(conn, name), "\n",
    "FIELDS TERMINATED BY ", dbQuoteString(conn, sep), "\n",
    "OPTIONALLY ENCLOSED BY ", dbQuoteString(conn, quote), "\n",
    "LINES TERMINATED BY ", dbQuoteString(conn, eol), "\n",
    "IGNORE ", skip + as.integer(header), " LINES"
  )

  dbExecute(conn, sql)

  invisible(TRUE)
}

#' @rdname mariadb-tables
#' @export
setMethod("dbWriteTable", c("MariaDBConnection", "character", "character"), dbWriteTable_MariaDBConnection_character_character)

#' @rdname transactions
#' @usage NULL
dbRollback_MariaDBConnection <- function(conn, ...) {
  connection_rollback(conn@ptr)
  invisible(TRUE)
}

#' @rdname transactions
#' @export
setMethod("dbRollback", "MariaDBConnection", dbRollback_MariaDBConnection)

#' Class MariaDBConnection.
#'
#' `"MariaDBConnection"` objects are usually created by [DBI::dbConnect()].
#' They represent a connection to a MariaDB or MySQL database.
#'
#' The `"MySQLConnection"` class is a subclass of `"MariaDBConnection"`.
#' Objects of that class are created by `dbConnect(MariaDB(), ..., mysql = TRUE)`
#' to indicate that the server is a MySQL server.
#' The \pkg{RMariaDB} package supports both MariaDB and MySQL servers, but the SQL dialect
#' and other details vary.
#' The default is to detect the server type based on the version number.
#'
#' The older \pkg{RMySQL} package also implements the `"MySQLConnection"` class.
#' The S4 system is able to distinguish between \pkg{RMariaDB} and \pkg{RMySQL} objects
#' even if both packages are loaded.
#'
#' @keywords internal
MariaDBConnection <- setClass("MariaDBConnection",
  contains = "DBIConnection",
  slots = list(
    ptr = "externalptr",
    host = "character",
    db = "character",
    load_data_local_infile = "logical",
    bigint = "character",
    timezone = "character",
    timezone_out = "character"
  )
)

#' @exportClass MariaDBConnection
NULL

#' @keywords internal
#' @name MariaDBConnection-class
#' @aliases MySQLConnection-class
MySQLConnection <- setClass("MySQLConnection",
  contains = "MariaDBConnection"
)

#' @exportClass MySQLConnection
NULL

# format()
#' @export
#' @rdname MariaDBConnection-class
format.MariaDBConnection <- function(x, ...) {
  if (dbIsValid(x)) {
    info <- dbGetInfo(x)
    details <- paste0(info$dbname, "@", info$host)
  } else {
    details <- "DISCONNECTED"
  }

  paste0("<MariaDBConnection> ", details)
}

#' @inheritParams DBI::sqlRownamesToColumn
#' @param overwrite a logical specifying whether to overwrite an existing table
#'   or not. Its default is `FALSE`.
#' @param append a logical specifying whether to append to an existing table
#'   in the DBMS.  If appending, then the table (or temporary table)
#'   must exist, otherwise an error is reported. Its default is `FALSE`.
#' @param value A data frame.
#' @param field.types Optional, overrides default choices of field types,
#'   derived from the classes of the columns in the data frame.
#' @param temporary If `TRUE`, creates a temporary table that expires
#'   when the connection is closed. For `dbRemoveTable()`, only temporary
#'   tables are considered if this argument is set to `TRUE`.
#' @rdname mariadb-tables
#' @usage NULL
dbWriteTable_MariaDBConnection_character_data.frame <- function(conn, name, value, field.types = NULL, row.names = FALSE,
                                                                overwrite = FALSE, append = FALSE, ...,
                                                                temporary = FALSE) {
  if (!is.data.frame(value)) {
    stopc("`value` must be data frame")
  }

  row.names <- compatRowNames(row.names)

  if ((!is.logical(row.names) && !is.character(row.names)) || length(row.names) != 1L) {
    stopc("`row.names` must be a logical scalar or a string")
  }
  if (!is.logical(overwrite) || length(overwrite) != 1L || is.na(overwrite)) {
    stopc("`overwrite` must be a logical scalar")
  }
  if (!is.logical(append) || length(append) != 1L || is.na(append)) {
    stopc("`append` must be a logical scalar")
  }
  if (!is.logical(temporary) || length(temporary) != 1L) {
    stopc("`temporary` must be a logical scalar")
  }
  if (overwrite && append) {
    stopc("overwrite and append cannot both be TRUE")
  }
  if (!is.null(field.types) && !(is.character(field.types) && !is.null(names(field.types)) && !anyDuplicated(names(field.types)))) {
    stopc("`field.types` must be a named character vector with unique names, or NULL")
  }
  if (append && !is.null(field.types)) {
    stopc("Cannot specify `field.types` with `append = TRUE`")
  }

  need_transaction <- !connection_is_transacting(conn@ptr)
  if (need_transaction) {
    dbBegin(conn)
    on.exit(dbRollback(conn))
  }

  if (!temporary) {
    found <- dbExistsTable(conn, name)
    if (found && !overwrite && !append) {
      stopc(
        "Table ", name, " exists in database, and both overwrite and",
        " append are FALSE"
      )
    }
  } else {
    found <- FALSE
  }

  if (overwrite) {
    dbRemoveTable(conn, name, temporary = temporary, fail_if_missing = FALSE)
  }

  # dbAppendTable() calls sql_data(), we only need to take care of row names
  row.names <- compatRowNames(row.names)
  value <- sqlRownamesToColumn(value, row.names)
  value <- factor_to_string(value)

  if (!found || overwrite) {
    if (is.null(field.types)) {
      combined_field_types <- lapply(value, dbDataType, dbObj = conn)
    } else {
      combined_field_types <- rep("", length(value))
      names(combined_field_types) <- names(value)
      field_types_idx <- match(names(field.types), names(combined_field_types))
      stopifnot(!any(is.na(field_types_idx)))
      combined_field_types[field_types_idx] <- field.types
      values_idx <- setdiff(seq_along(value), field_types_idx)
      combined_field_types[values_idx] <- lapply(value[values_idx], dbDataType, dbObj = conn)
    }

    dbCreateTable(
      conn = conn,
      name = name,
      fields = combined_field_types,
      temporary = temporary
    )
  }

  if (nrow(value) > 0) {
    if (conn@load_data_local_infile) {
      out <- db_append_table(
        conn = conn,
        name = name,
        value = value,
        warn_factor = FALSE,
        safe = TRUE,
        transact = FALSE
      )
    } else {
      out <- dbAppendTable(
        conn = conn,
        name = name,
        value = value
      )
    }

    if (out < nrow(value)) {
      msg <- paste0("Error writing table: sent ", nrow(value), " rows, added ", out, " rows.")
      if (need_transaction) {
        stopc(msg)
      } else {
        warningc(msg)
      }
    }
  }

  if (need_transaction) {
    on.exit(NULL)
    dbCommit(conn)
  }

  invisible(TRUE)
}

#' @rdname mariadb-tables
#' @export
setMethod("dbWriteTable", c("MariaDBConnection", "character", "data.frame"), dbWriteTable_MariaDBConnection_character_data.frame)

#' @rdname MariaDBDriver-class
#' @usage NULL
dbUnloadDriver_MariaDBDriver <- function(drv, ...) {
  TRUE
}

#' @rdname MariaDBDriver-class
setMethod("dbUnloadDriver", "MariaDBDriver", dbUnloadDriver_MariaDBDriver)

# show()
#' @rdname MariaDBConnection-class
#' @usage NULL
show_MariaDBConnection <- function(object) {
  info <- dbGetInfo(object)
  cat("<", class(object), ">\n", sep = "")
  if (dbIsValid(object)) {
    conInfo <- paste0(info$user, "@", info$host)
    if (info$port != 3306) {
      conInfo <- paste0(conInfo, ":", info$port)
    }
    conInfo <- paste0(conInfo, "<", info$dbname, ">[", info$thread.id, "]")
    conParts <- unlist(strsplit(info$con.type, split = " ", fixed = T))
    conInfo <- paste(conInfo, paste(conParts[-1], collapse = " "))
    if (!is.null(info$ssl.cipher)) {
      conInfo <- paste(conInfo, "over SSL")
    }
    cat("  Connection: ", conInfo, "\n", sep = "")
  } else {
    cat("  DISCONNECTED\n")
  }
}

#' @rdname MariaDBConnection-class
#' @export
setMethod("show", "MariaDBConnection", show_MariaDBConnection)

#' @name mariadb-tables
#' @details
#' When using `load_data_local_infile = TRUE` in [dbConnect()],
#' pass `safe = FALSE` to `dbAppendTable()` to avoid transactions.
#' Because `LOAD DATA INFILE` is used internally, this means that
#' rows violating primary key constraints are now silently ignored.
#' @importFrom utils write.table
#' @usage NULL
dbAppendTable_MariaDBConnection <- function(conn, name, value, ..., row.names = NULL) {
  if (!is.null(row.names)) {
    stop("Can't pass `row.names` to `dbAppendTable()`", call. = FALSE)
  }
  stopifnot(is.data.frame(value))

  if (!conn@load_data_local_infile) {
    return(callNextMethod())
  }

  db_append_table(conn, name, value, ..., warn_factor = TRUE, transact = TRUE)
}

#' @rdname mariadb-tables
#' @export
setMethod("dbAppendTable", "MariaDBConnection", dbAppendTable_MariaDBConnection)

# dbDisconnect()
#' @rdname MariaDBConnection-class
#' @usage NULL
dbDisconnect_MariaDBConnection <- function(conn, ...) {
  connection_release(conn@ptr)
  invisible(TRUE)
}

#' @rdname MariaDBConnection-class
#' @export
setMethod("dbDisconnect", "MariaDBConnection", dbDisconnect_MariaDBConnection)

#' @include MariaDBConnection.R
NULL

#' Quote MariaDB strings and identifiers.
#'
#' In MariaDB, identifiers are enclosed in backticks, e.g. `` `x` ``.
#'
#' @keywords internal
#' @name mariadb-quoting
#' @examples
#' if (mariadbHasDefault()) {
#'   con <- dbConnect(RMariaDB::MariaDB())
#'   dbQuoteIdentifier(con, c("a b", "a`b"))
#'   dbQuoteString(con, c("a b", "a'b"))
#'   dbDisconnect(con)
#' }
NULL

as_table <- function(schema, table) {
  args <- c(schema = schema, table = table)
  # Also omits NA args
  args <- args[!is.na(args) & args != ""]
  do.call(Id, as.list(args))
}

cast_difftime <- function(x) {
  SQL(paste0("CAST(", x, " AS time)"))
}

#' @include MariaDBConnection.R
NULL

#' Read and write MariaDB tables.
#'
#' These methods read or write entire tables from a MariaDB database.
#'
#' @return A data.frame in the case of `dbReadTable()`; otherwise a logical
#' indicating whether the operation was successful.
#' @note The data.frame returned by `dbReadTable()` only has
#' primitive data, e.g., it does not coerce character data to factors.
#' Temporary tables are ignored for `dbExistsTable()` and `dbListTables()` due to
#' limitations of the underlying C API. For this reason, a prior existence check
#' is performed only before creating a regular persistent table; an attempt to
#' create a temporary table with an already existing name will fail with a
#' message from the database driver.
#'
#'
#' @param conn a [MariaDBConnection-class] object, produced by
#'   [DBI::dbConnect()]
#' @param name a character string specifying a table name.
#' @param check.names If `TRUE`, the default, column names will be
#'   converted to valid R identifiers.
#' @inheritParams DBI::sqlRownamesToColumn
#' @param ... Unused, needed for compatibility with generic.
#' @examples
#' if (mariadbHasDefault()) {
#'   con <- dbConnect(RMariaDB::MariaDB(), dbname = "test")
#'
#'   # By default, row names are written in a column to row_names, and
#'   # automatically read back into the row.names()
#'   dbWriteTable(con, "mtcars", mtcars[1:5, ], temporary = TRUE)
#'   dbReadTable(con, "mtcars")
#'   dbReadTable(con, "mtcars", row.names = FALSE)
#' }
#' @name mariadb-tables
NULL

db_append_table <- function(conn, name, value, warn_factor = TRUE, safe = TRUE, transact = TRUE) {
  path <- tempfile("RMariaDB", fileext = ".tsv")
  is_list <- vlapply(value, is.list)
  colnames <- dbQuoteIdentifier(conn, names(value))
  if (any(is_list)) {
    set <- paste0(
      "SET ",
      paste0(
        colnames[is_list], " = UNHEX(@X", which(is_list), ")",
        collapse = ", "
      )
    )
    colnames[is_list] <- paste0("@X", which(is_list))
  } else {
    set <- ""
  }

  quoted_name <- dbQuoteIdentifier(conn, name)
  if (length(quoted_name) != 1) {
    stop("Must pass one table name", call. = FALSE)
  }

  sql <- paste0(
    "LOAD DATA LOCAL INFILE ", dbQuoteString(conn, path), "\n",
    "IGNORE\n",
    "INTO TABLE ", quoted_name, "\n",
    "CHARACTER SET utf8mb4 \n",
    "(", paste0(colnames, collapse = ", "), ")",
    set
  )

  file <- file(path, "wb")
  on.exit(close(file))

  readr::write_delim(
    csv_quote(value, warn_factor, conn), file, quote = "none", delim = "\t", na = "\\N",
    col_names = FALSE
  )

  # Close connection manually, unlink when done to save disk space
  on.exit(unlink(path), add = FALSE)
  close(file)

  if (safe) {
    if (transact) {
      dbBegin(conn)
      on.exit(dbRollback(conn), add = TRUE)
    }

    sql_count <- paste0("SELECT COUNT(*) FROM ", quoted_name)
    count_before <- dbGetQuery(conn, sql_count)[[1]]

    out <- dbExecute(conn, sql)

    # Some servers don't return a record count here,
    # need to count manually
    if (out == 0) {
      count_after <- dbGetQuery(conn, sql_count)[[1]]
      out <- as.numeric(count_after - count_before)
    }

    if (transact) {
      if (out < nrow(value)) {
        stopc("Error writing table: sent ", nrow(value), " rows, added ", out, " rows.")
      }
      dbCommit(conn)
    }

    # Manual cleanup
    unlink(path)
    on.exit(NULL, add = FALSE)

    out
  } else {
    dbExecute(conn, sql)
  }
}

csv_quote <- function(x, warn_factor, conn) {
  old <- options(digits.secs = 6)
  on.exit(options(old))

  x[] <- lapply(x, csv_quote_one, conn)
  factor_to_string(x, warn = warn_factor)
}

csv_quote_one <- function(x, conn) {
  if (inherits(x, "AsIs")) {
    class(x) <- setdiff(class(x), "AsIs")
  }

  if (is.factor(x)) {
    levels(x) <- csv_quote_char(levels(x))
  } else if (is.character(x)) {
    x <- csv_quote_char(x)
  } else if (is.integer(x)) {
    x_orig <- x
    x <- as.character(x)
    # Failure to load BIT(1) columns with a verbatim 0 (???)
    # https://stackoverflow.com/a/17836602/946850
    x[!is.na(x_orig) & x_orig == 0] <- ""
  } else if (is.integer64(x)) {
    x_orig <- x
    x <- as.character(x)
    # Failure to load BIT(1) columns with a verbatim 0 (???)
    # https://stackoverflow.com/a/17836602/946850
    x[!is.na(x_orig) & x_orig == 0] <- ""
  } else if (is.numeric(x)) {
    x_orig <- x
    if (all_integerish(x)) {
      x <- formatC(x, format = "d")
    } else {
      # https://dev.mysql.com/doc/refman/5.7/en/number-literals.html
      x <- formatC(x, digits = 17, format = "E")
    }
    x[is.na(x_orig) | is.infinite(x_orig)] <- NA_character_
  } else if (is.logical(x)) {
    x <- as.character(as.integer(x))
  } else if (inherits(x, "Date")) {
    x <- as.character(x)
  } else if (inherits(x, "difftime")) {
    x <- hms::as_hms(x)
    x <- as.character(x)
  } else if (inherits(x, "POSIXt")) {
    x <- format(x, format = "%Y-%m-%dT%H:%M:%OS", tz = conn@timezone)
  } else if (inherits(x, "list")) {
    x_orig <- x
    x <- vcapply(x, function(x) paste(format(x), collapse = ""))
    x[vlapply(x_orig, is.null)] <- NA_character_
  } else {
    stop("NYI: ", paste(class(x), collapse = "/"), call. = FALSE)
  }

  x
}

csv_quote_char <- function(x) {
  x <- enc2utf8(x)
  x <- gsub("\\", "\\\\", x, fixed = TRUE)
  x <- gsub("\t", "\\t", x, fixed = TRUE)
  x <- gsub("\r", "\\r", x, fixed = TRUE)
  x <- gsub("\n", "\\n", x, fixed = TRUE)
  x
}

all_integerish <- function(x) {
  x <- x[!is.na(x)]
  if (any(is.infinite(x))) {
    return(FALSE)
  }
  all(x >= -2147483647) && all(x <= 2147483647) && all(x == as.integer(x))
}

get_char_type <- function(x) {
  width <- max(nchar(enc2utf8(x)), 1, na.rm = TRUE)
  if (width > 255) {
    "TEXT"
  } else {
    paste0("VARCHAR(", width, ")")
  }
}

#' @rdname dbDataType
#' @usage NULL
dbDataType_MariaDBDriver <- function(dbObj, obj, ...) {
  if (is.factor(obj)) return(get_char_type(levels(obj)))
  if (inherits(obj, "POSIXct")) return("DATETIME(6)")
  if (inherits(obj, "Date")) return("DATE")
  if (inherits(obj, "difftime")) return("TIME(6)")
  if (inherits(obj, "integer64")) return("BIGINT")
  if (is.data.frame(obj)) return(callNextMethod(dbObj, obj))

  switch(typeof(obj),
    logical = "TINYINT", # works better than BIT(1), https://stackoverflow.com/q/289727/946850
    integer = "INTEGER",
    double = "DOUBLE",
    character = get_char_type(obj),
    list = "BLOB",
    stop("Unsupported type", call. = FALSE)
  )
}

#' @rdname dbDataType
#' @export
setMethod("dbDataType", "MariaDBDriver", dbDataType_MariaDBDriver)

if (rlang::is_installed("DBItest")) DBItest::make_context(
  MariaDB(),
  c(
    mariadb_default_args,
    list(load_data_local_infile = (Sys.getenv("RMARIADB_LOAD_DATA_LOCAL_INFILE") != "") && rlang::is_installed("readr"))
  ),
  tweaks = DBItest::tweaks(
    dbitest_version = "1.7.2",
    constructor_relax_args = TRUE,
    placeholder_pattern = "?",
    logical_return = function(x) as.integer(x),
    list_temporary_tables = FALSE
  ),
  name = "RMariaDB",
  default_skip = c(
    # result
    "data_logical",                               # not an error: cannot cast to logical
    "data_raw",                                   # not an error: can't cast to blob type

    # Fails on Ubuntu 18.04:
    "list_objects_features",

    # bad tests
    if ((.Platform$OS.type == "windows" && .Platform$r_arch == "i386") || packageVersion("DBItest") < "1.7.2") "append_roundtrip_timestamp",
    if ((.Platform$OS.type == "windows" && .Platform$r_arch == "i386") || packageVersion("DBItest") < "1.7.2") "roundtrip_timestamp",

    NULL
  )
)

#' @include MariaDBConnection.R
NULL

#' DBMS Transaction Management
#'
#' Commits or roll backs the current transaction in an MariaDB connection.
#' Note that in MariaDB DDL statements (e.g. `CREATE TABLE`) cannot
#' be rolled back.
#'
#' @param conn a [MariaDBConnection-class] object, as produced by
#'   [DBI::dbConnect()].
#' @param ... Unused.
#' @examples
#' if (mariadbHasDefault()) {
#'   con <- dbConnect(RMariaDB::MariaDB(), dbname = "test")
#'   df <- data.frame(id = 1:5)
#'
#'   dbWriteTable(con, "df", df, temporary = TRUE)
#'   dbBegin(con)
#'   dbExecute(con, "UPDATE df SET id = id * 10")
#'   dbGetQuery(con, "SELECT id FROM df")
#'   dbRollback(con)
#'
#'   dbGetQuery(con, "SELECT id FROM df")
#'
#'   dbDisconnect(con)
#' }
#' @name transactions
NULL

#' @rdname mariadb-quoting
#' @usage NULL
dbQuoteLiteral_MariaDBConnection <- function(conn, x, ...) {
  # Switchpatching to avoid ambiguous S4 dispatch, so that our method
  # is used only if no alternatives are available.

  if (inherits(x, "difftime")) return(cast_difftime(callNextMethod()))

  callNextMethod()
}

#' @rdname mariadb-quoting
#' @export
setMethod("dbQuoteLiteral", signature("MariaDBConnection"), dbQuoteLiteral_MariaDBConnection)

#' @rdname mariadb-quoting
#' @usage NULL
dbQuoteString_MariaDBConnection_character <- function(conn, x, ...) {
  SQL(connection_quote_string(conn@ptr, enc2utf8(x)))
}

#' @rdname mariadb-quoting
#' @export
setMethod("dbQuoteString", c("MariaDBConnection", "character"), dbQuoteString_MariaDBConnection_character)

#' @inheritParams DBI::dbListObjects
#' @rdname mariadb-tables
#' @usage NULL
dbListObjects_MariaDBConnection_ANY <- function(conn, prefix = NULL, ...) {
  query <- NULL
  if (is.null(prefix)) {
    # DATABASE(): https://stackoverflow.com/a/8096574/946850
    query <- paste0(
      "SELECT NULL AS `schema`, table_name AS `table` FROM INFORMATION_SCHEMA.tables\n",
      "WHERE table_schema = DATABASE()\n",
      "UNION ALL\n",
      "SELECT DISTINCT table_schema AS `schema`, NULL AS `table` FROM INFORMATION_SCHEMA.tables"
    )
  } else {
    unquoted <- dbUnquoteIdentifier(conn, prefix)
    is_prefix <- vlapply(unquoted, function(x) {
      "schema" %in% names(x@name) && !("table" %in% names(x@name))
    })
    schemas <- vcapply(unquoted[is_prefix], function(x) x@name[["schema"]])
    if (length(schemas) > 0) {
      schema_strings <- dbQuoteString(conn, schemas)
      query <- paste0(
        "SELECT table_schema AS `schema`, table_name AS `table` FROM INFORMATION_SCHEMA.tables\n",
        "WHERE ",
        "(table_schema IN (", paste(schema_strings, collapse = ", "), "))"
      )
    }
  }

  if (is.null(query)) {
    res <- data.frame(schema = character(), table = character(), stringsAsFactors = FALSE)
  } else {
    res <- dbGetQuery(conn, query)
  }

  is_prefix <- !is.na(res$schema) & is.na(res$table)
  tables <- Map(res$schema, res$table, f = as_table)

  ret <- data.frame(
    table = I(unname(tables)),
    is_prefix = is_prefix,
    stringsAsFactors = FALSE
  )
  ret
}

#' @rdname mariadb-tables
#' @export
setMethod("dbListObjects", c("MariaDBConnection", "ANY"), dbListObjects_MariaDBConnection_ANY)

#' Class MariaDBDriver with constructor MariaDB.
#'
#' An MariaDB driver implementing the R database (DBI) API.
#' This class should always be initialized with the [MariaDB()] function.
#' It returns a singleton that allows you to connect to MariaDB.
#'
#' @export
#' @keywords internal
setClass("MariaDBDriver",
  contains = "DBIDriver",
)

#' MariaDB Check for Compiled Versus Loaded Client Library Versions
#'
#' This function prints out the compiled and loaded client library versions.
#'
#' @return A named integer vector of length two, the first element
#'   representing the compiled library version and the second element
#'   representing the loaded client library version.
#' @export
#' @examples
#' mariadbClientLibraryVersions()
mariadbClientLibraryVersions <- function() {
  version()
}

# Set during installation time for the correct library
PACKAGE_VERSION <- utils::packageVersion(utils::packageName())

# Created with:
# methods::getGenerics(asNamespace("DBI")) %>%
#   grep("^db[A-Z]", ., value = TRUE) %>%
#   setdiff(c("dbCallProc", "dbListConnections", "dbSetDataMappings", "dbGetException")) %>%
#   paste0("#' @exportMethod ", ., "\nNULL\n", collapse = "\n") %>%
#   cat(file = "R/export.R")
#
#' @exportMethod dbAppendTable
NULL

#' @exportMethod dbBegin
NULL

#' @exportMethod dbBind
NULL

#' @exportMethod dbCanConnect
NULL

#' @exportMethod dbClearResult
NULL

#' @exportMethod dbColumnInfo
NULL

#' @exportMethod dbCommit
NULL

#' @exportMethod dbConnect
NULL

#' @exportMethod dbCreateTable
NULL

#' @exportMethod dbDataType
NULL

#' @exportMethod dbDisconnect
NULL

#' @exportMethod dbDriver
NULL

#' @exportMethod dbExecute
NULL

#' @exportMethod dbExistsTable
NULL

#' @exportMethod dbFetch
NULL

#' @exportMethod dbGetInfo
NULL

#' @exportMethod dbGetQuery
NULL

#' @exportMethod dbGetRowCount
NULL

#' @exportMethod dbGetRowsAffected
NULL

#' @exportMethod dbGetStatement
NULL

#' @exportMethod dbHasCompleted
NULL

#' @exportMethod dbIsReadOnly
NULL

#' @exportMethod dbIsValid
NULL

#' @exportMethod dbListFields
NULL

#' @exportMethod dbListObjects
NULL

#' @exportMethod dbListResults
NULL

#' @exportMethod dbListTables
NULL

#' @exportMethod dbQuoteIdentifier
NULL

#' @exportMethod dbQuoteLiteral
NULL

#' @exportMethod dbQuoteLiteral
NULL

#' @exportMethod dbQuoteString
NULL

#' @exportMethod dbReadTable
NULL

#' @exportMethod dbRemoveTable
NULL

#' @exportMethod dbRollback
NULL

#' @exportMethod dbSendQuery
NULL

#' @exportMethod dbSendStatement
NULL

#' @exportMethod dbUnloadDriver
NULL

#' @exportMethod dbUnquoteIdentifier
NULL

#' @exportMethod dbWithTransaction
NULL

#' @exportMethod dbWriteTable
NULL

#' @export
DBI::Id

astyle <- function(extra_args = character()) {
  astyle_cmd <- "astyle"
  if (Sys.which(astyle_cmd) == "") {
    skip("astyle not found")
  }

  astyle_args <- c(
    "-n",
    "--indent=spaces=2",
    "--indent-cases",
    "--unpad-paren",
    "--pad-header",
    "--pad-oper",
    "--min-conditional-indent=0",
    "--align-pointer=type",
    "--align-reference=type"
  )

  tryCatch(
    {
      src_path <- testthat::test_path("../../src")
      src_path <- normalizePath(src_path)
    },
    warning = function(e) {
      skip(paste0("Sources not found: ", conditionMessage(e)))
    },
    error = function(e) {
      skip(paste0("Sources not found: ", conditionMessage(e)))
    }
  )
  src_files <- dir(src_path, "[.](?:cpp|h)$", recursive = FALSE, full.names = TRUE)
  astyle_files <- grep("(?:RcppExports[.]cpp)", src_files, value = TRUE, invert = TRUE)
  output <- system2(astyle_cmd, c(astyle_args, astyle_files, extra_args), stdout = TRUE, stderr = TRUE)
  unchanged <- grepl("^Unchanged", output)
  if (any(!unchanged)) {
    warning(paste(output[!unchanged], collapse = "\n"))
  }
}

#' @name mariadb-quoting
#' @usage NULL
dbQuoteIdentifier_MariaDBConnection_Id <- function(conn, x, ...) {
  unqouted_name <- x@name
  if (length(unqouted_name) == 3 && unqouted_name[[1]] == "def") {
    unqouted_name <- unqouted_name[-1]
  }

  SQL(paste0(dbQuoteIdentifier(conn, unqouted_name), collapse = "."))
}

#' @rdname mariadb-quoting
#' @export
setMethod("dbQuoteIdentifier", c("MariaDBConnection", "Id"), dbQuoteIdentifier_MariaDBConnection_Id)

# dbIsValid()
#' @rdname MariaDBConnection-class
#' @usage NULL
dbIsValid_MariaDBConnection <- function(dbObj, ...) {
  connection_valid(dbObj@ptr)
}

#' @rdname MariaDBConnection-class
#' @export
setMethod("dbIsValid", "MariaDBConnection", dbIsValid_MariaDBConnection)

#' @name transactions
#' @usage NULL
dbBegin_MariaDBConnection <- function(conn, ...) {
  connection_begin_transaction(conn@ptr)
  invisible(TRUE)
}

#' @rdname transactions
#' @export
setMethod("dbBegin", "MariaDBConnection", dbBegin_MariaDBConnection)

#' @rdname query
#' @usage NULL
dbGetStatement_MariaDBResult <- function(res, ...) {
  if (!dbIsValid(res)) {
    stopc("Expired, result set already closed")
  }
  res@sql
}

#' @rdname query
#' @export
setMethod("dbGetStatement", "MariaDBResult", dbGetStatement_MariaDBResult)

#' @name query
#' @usage NULL
dbBind_MariaDBResult <- function(res, params, ...) {
  if (!is.null(names(params))) {
    stopc("Cannot use named parameters for anonymous placeholders")
  }

  params <- sql_data(as.list(params), res@conn, warn = TRUE)

  result_bind(res@ptr, params)
  invisible(res)
}

#' @rdname query
#' @export
setMethod("dbBind", "MariaDBResult", dbBind_MariaDBResult)

#' @rdname MariaDBDriver-class
#' @usage NULL
dbGetInfo_MariaDBDriver <- function(dbObj, ...) {
  client_version <- names(version())[[2]]

  list(
    driver.version = PACKAGE_VERSION,
    client.version = package_version(client_version)
  )
}

#' @rdname MariaDBDriver-class
#' @export
setMethod("dbGetInfo", "MariaDBDriver", dbGetInfo_MariaDBDriver)

#' @rdname transactions
#' @usage NULL
dbCommit_MariaDBConnection <- function(conn, ...) {
  connection_commit(conn@ptr)
  invisible(TRUE)
}

#' @rdname transactions
#' @export
setMethod("dbCommit", "MariaDBConnection", dbCommit_MariaDBConnection)

#' Execute a SQL statement on a database connection.
#'
#' To retrieve results a chunk at a time, use [dbSendQuery()],
#' [dbFetch()], then [dbClearResult()]. Alternatively, if you want all the
#' results (and they'll fit in memory) use [dbGetQuery()] which sends,
#' fetches and clears for you. For data manipulation queries (i.e. queries
#' that do not return data, such as \code{UPDATE}, \code{DELETE}, etc.),
#' [dbSendStatement()] serves as a counterpart to [dbSendQuery()], while
#' [dbExecute()] corresponds to [dbGetQuery()].
#'
#' @param conn A [MariaDBConnection-class] object.
#' @param res A [MariaDBResult-class] object.
#' @inheritParams DBI::sqlRownamesToColumn
#' @param n Number of rows to retrieve. Use -1 to retrieve all rows.
#' @param params A list of query parameters to be substituted into
#'   a parameterised query.
#' @param statement A character vector of length one specifying the SQL
#'   statement that should be executed.  Only a single SQL statement should be
#'   provided.
#' @param ... Unused. Needed for compatibility with generic.
#' @param immediate If TRUE, uses the `mysql_real_query()` API
#'   instead of `mysql_stmt_init()`.
#'   This allows passing multiple statements (with [CLIENT_MULTI_STATEMENTS])
#'   and turns off the ability to pass parameters.
#' @examples
#' if (mariadbHasDefault()) {
#'   con <- dbConnect(RMariaDB::MariaDB(), dbname = "test")
#'   dbWriteTable(con, "arrests", datasets::USArrests, temporary = TRUE)
#'
#'   # Run query to get results as dataframe
#'   dbGetQuery(con, "SELECT * FROM arrests limit 3")
#'
#'   # Send query to pull requests in batches
#'   res <- dbSendQuery(con, "SELECT * FROM arrests")
#'   data <- dbFetch(res, n = 2)
#'   data
#'   dbHasCompleted(res)
#'
#'   dbClearResult(res)
#'   dbDisconnect(con)
#' }
#' @rdname query
#' @usage NULL
dbFetch_MariaDBResult <- function(res, n = -1, ..., row.names = FALSE) {
  if (length(n) != 1) stopc("n must be scalar")
  if (n < -1) stopc("n must be nonnegative or -1")
  if (is.infinite(n)) n <- -1
  if (trunc(n) != n) stopc("n must be a whole number")
  ret <- result_fetch(res@ptr, n = n)
  ret <- convert_bigint(ret, res@bigint)
  ret <- fix_timezone(ret, res@conn)
  ret <- fix_blob(ret)
  ret <- sqlColumnToRownames(ret, row.names)
  ret
}

#' @rdname query
#' @export
setMethod("dbFetch", "MariaDBResult", dbFetch_MariaDBResult)

#' @rdname mariadb-tables
#' @usage NULL
dbListTables_MariaDBConnection <- function(conn, ...) {
  # DATABASE(): https://stackoverflow.com/a/8096574/946850
  dbGetQuery(
    conn,
    paste0(
      "SELECT table_name FROM INFORMATION_SCHEMA.tables\n",
      "WHERE table_schema = DATABASE()"
    )
  )[[1]]
}

#' @rdname mariadb-tables
#' @export
setMethod("dbListTables", "MariaDBConnection", dbListTables_MariaDBConnection)

#' Determine the SQL Data Type of an S object
#'
#' This method is a straight-forward implementation of the corresponding
#' generic function.
#'
#' @param dbObj A [MariaDBDriver-class] or [MariaDBConnection-class] object.
#' @param obj R/S-Plus object whose SQL type we want to determine.
#' @param \dots any other parameters that individual methods may need.
#' @name dbDataType
#' @examples
#' dbDataType(RMariaDB::MariaDB(), "a")
#' dbDataType(RMariaDB::MariaDB(), 1:3)
#' dbDataType(RMariaDB::MariaDB(), 2.5)
#' @usage NULL
dbDataType_MariaDBConnection <- function(dbObj, obj, ...) {
  dbDataType(MariaDB(), obj, ...)
}

#' @rdname dbDataType
#' @export
setMethod("dbDataType", "MariaDBConnection", dbDataType_MariaDBConnection)

#' @rdname result-meta
#' @usage NULL
dbHasCompleted_MariaDBResult <- function(res, ...) {
  result_has_completed(res@ptr)
}

#' @rdname result-meta
#' @export
setMethod("dbHasCompleted", "MariaDBResult", dbHasCompleted_MariaDBResult)

#' @rdname query
#' @usage NULL
dbSendQuery_MariaDBConnection_character <- function(conn, statement, params = NULL, ..., immediate = FALSE) {
  # immediate = TRUE not supported
  # Can't raise a warning here due to a test in DBItest
  dbSend(conn, statement, params, is_statement = FALSE, immediate = FALSE)
}

#' @rdname query
#' @export
setMethod("dbSendQuery", c("MariaDBConnection", "character"), dbSendQuery_MariaDBConnection_character)

#' Connect/disconnect to a MariaDB DBMS
#'
#' These methods are straight-forward implementations of the corresponding
#' generic functions.
#'
#' @section Time zones:
#' MySQL and MariaDB support named time zones,
#' they must be installed on the server.
#' See <https://dev.mysql.com/doc/mysql-g11n-excerpt/8.0/en/time-zone-support.html>
#' for more details.
#' Without installation, time zone support is restricted to UTC offset,
#' which cannot take into account DST offsets.
#'
#' @section Secure passwords:
#' Avoid storing passwords hard-coded in the code, use e.g. the \pkg{keyring}
#' package to store and retrieve passwords in a secure way.
#'
#' The MySQL client library (but not MariaDB) supports a `.mylogin.cnf` file
#' that can be passed in the `default.file` argument.
#' This file can contain an obfuscated password, which is not a secure way
#' to store passwords but may be acceptable if the user is aware of the
#' restrictions.
#' The availability of this feature depends on the client library used
#' for compiling the \pkg{RMariaDB} package.
#' Windows and macOS binaries on CRAN are compiled against the MariaDB Connector/C
#' client library which do not support this feature.
#'
#' @param drv an object of class [MariaDBDriver-class] or
#'   [MariaDBConnection-class].
#' @param username,password Username and password. If username omitted,
#'   defaults to the current user. If password is omitted, only users
#'   without a password can log in.
#' @param dbname string with the database name or NULL. If not NULL, the
#'   connection sets the default database to this value.
#' @param host string identifying the host machine running the MariaDB server or
#'   NULL. If NULL or the string `"localhost"`, a connection to the local
#'   host is assumed.
#' @param unix.socket (optional) string of the unix socket or named pipe.
#' @param port (optional) integer of the TCP/IP default port.
#' @param client.flag (optional) integer setting various MariaDB client flags,
#'   see [Client-flags] for details.
#' @param group string identifying a section in the `default.file` to use
#'   for setting authentication parameters (see [MariaDB()]).
#' @param groups deprecated, use `group` instead.
#' @param default.file string of the filename with MariaDB client options,
#'   only relevant if `groups` is given. The default value depends on the
#'   operating system (see references), on Linux and OS X the files
#'   `~/.my.cnf` and `~/.mylogin.cnf` are used. Expanded with [normalizePath()].
#' @param ssl.key (optional) string of the filename of the SSL key file to use.
#'   Expanded with [normalizePath()].
#' @param ssl.cert (optional) string of the filename of the SSL certificate to
#'   use. Expanded with [normalizePath()].
#' @param ssl.ca (optional) string of the filename of an SSL certificate
#'   authority file to use. Expanded with [normalizePath()].
#' @param ssl.capath (optional) string of the path to a directory containing
#'   the trusted SSL CA certificates in PEM format. Expanded with
#'   [normalizePath()].
#' @param ssl.cipher (optional) string list of permitted ciphers to use for SSL
#'   encryption.
#' @param ... Unused, needed for compatibility with generic.
#' @param load_data_local_infile Set to `TRUE` to use `LOAD DATA LOCAL INFILE`
#'   in [dbWriteTable()] and [dbAppendTable()] by default.
#'   This capability is disabled by default on the server side
#'   for recent versions of MySQL Server.
#' @param bigint The R type that 64-bit integer types should be mapped to,
#'   default is [bit64::integer64], which allows the full range of 64 bit
#'   integers.
#' @param timeout Connection timeout, in seconds. Use `Inf` or a negative value
#'   for no timeout.
#' @param timezone (optional) time zone for the connection,
#'   the default corresponds to UTC.
#'   Set this argument if your server or database is configured with a different
#'   time zone than UTC.
#'   Set to `NULL` to automatically determine the server time zone.
#' @param timezone_out The time zone returned to R.
#'   The default is to use the value of the `timezone` argument,
#'   `"+00:00"` is converted to `"UTC"`
#'   If you want to display datetime values in the local timezone,
#'   set to [Sys.timezone()] or `""`.
#'   This setting does not change the time values returned, only their display.
#' @param reconnect (experimental) Set to `TRUE` to use `MYSQL_OPT_RECONNECT` to enable
#'   automatic reconnection. This is experimental and could be dangerous if the connection
#'   is lost in the middle of a transaction.
#' @param mysql Set to `TRUE`/`FALSE` to connect to a MySQL server or to a MariaDB server,
#'   respectively.
#'   The \pkg{RMariaDB} package supports both MariaDB and MySQL servers, but the SQL dialect
#'   and other details vary.
#'   The default is to assume MariaDB if the version is >= 10.0.0, and MySQL otherwise.
#'
#' @references
#' Configuration files: https://mariadb.com/kb/en/library/configuring-mariadb-with-mycnf/
#' @examples
#' \dontrun{
#' # Connect to a MariaDB database running locally
#' con <- dbConnect(RMariaDB::MariaDB(), dbname = "mydb")
#' # Connect to a remote database with username and password
#' con <- dbConnect(RMariaDB::MariaDB(),
#'   host = "mydb.mycompany.com",
#'   user = "abc", password = "def"
#' )
#' # But instead of supplying the username and password in code, it's usually
#' # better to set up a group in your .my.cnf (usually located in your home
#' # directory). Then it's less likely you'll inadvertently share them.
#' con <- dbConnect(RMariaDB::MariaDB(), group = "test")
#'
#' # To connect to a remote database and require the use of SSL
#' (and an example of using environment variables for your sensitive info)
#' con <- dbConnect(RMariaDB::MariaDB(),
#'   dbname = Sys.getenv('DB_NAME'),
#'   host = Sys.getenv('DB_HOST'),
#'   user = Sys.getenv('DB_USER'),
#'   password = Sys.getenv('DB_PASSWORD'),
#'   client.flag = CLIENT_SSL
#' )
#'
#' # Always cleanup by disconnecting the database
#' dbDisconnect(con)
#' }
#'
#' # All examples use the rs-dbi group by default.
#' if (mariadbHasDefault()) {
#'   con <- dbConnect(RMariaDB::MariaDB(), dbname = "test")
#'   con
#'   dbDisconnect(con)
#' }
#'
#' @usage NULL
#' @rdname dbConnect-MariaDBDriver-method
dbConnect_MariaDBDriver <- function(
    drv,
    dbname = NULL,
    username = NULL,
    password = NULL,
    host = NULL,
    unix.socket = NULL,
    port = 0,
    client.flag = 0,
    group = "rs-dbi",
    default.file = NULL,
    ssl.key = NULL,
    ssl.cert = NULL,
    ssl.ca = NULL,
    ssl.capath = NULL,
    ssl.cipher = NULL,
    ...,
    groups = NULL,
    load_data_local_infile = FALSE,
    bigint = c("integer64", "integer", "numeric", "character"),
    timeout = 10,
    timezone = "+00:00",
    timezone_out = NULL,
    reconnect = FALSE,
    mysql = NULL) {
  #
  bigint <- match.arg(bigint)

  if (is.infinite(timeout)) {
    timeout <- -1L
  } else {
    timeout <- as.integer(timeout)
  }

  # Make sure that `~` is resolved correctly:
  if (!is.null(default.file)) {
    default.file <- normalizePath(default.file)
  }
  if (!is.null(ssl.key)) {
    ssl.key <- normalizePath(ssl.key)
  }
  if (!is.null(ssl.cert)) {
    ssl.cert <- normalizePath(ssl.cert)
  }
  if (!is.null(ssl.ca)) {
    ssl.ca <- normalizePath(ssl.ca)
  }
  if (!is.null(ssl.capath)) {
    ssl.capath <- normalizePath(ssl.capath)
  }

  if (isTRUE(load_data_local_infile)) {
    if (!rlang::is_installed("readr")) {
      stopc("`load_data_local_infile = TRUE` requires the readr package.")
    }
  }

  if (!is.null(groups)) {
    warningc("Argument `groups` is deprecated, use `group` instead. ")
    if (missing(group)) {
      group <- groups
    }
  }

  reconnect <- isTRUE(reconnect)

  ptr <- connection_create(
    host, username, password, dbname, as.integer(port), unix.socket,
    as.integer(client.flag), group, default.file,
    ssl.key, ssl.cert, ssl.ca, ssl.capath, ssl.cipher,
    timeout, reconnect
  )

  info <- connection_info(ptr)

  if (is.null(mysql)) {
    if (info$db.version.int >= 100000) {
      mysql <- FALSE
    } else if (grepl("^5[.][0-9]+[.][0-9]+-.*mariadb", info$db.version, ignore.case = TRUE)) {
      # MariaDB built against MySQL libraries will report a 5.x version number
      # https://github.com/OpenNebula/one/issues/3924
      mysql <- FALSE
    } else {
      mysql <- TRUE
    }
  }

  if (isTRUE(mysql)) {
    new <- MySQLConnection
  } else {
    new <- MariaDBConnection
  }

  conn <- new(
    ptr = ptr,
    host = info$host,
    db = info$dbname,
    load_data_local_infile = isTRUE(load_data_local_infile),
    bigint = bigint
  )

  on.exit(dbDisconnect(conn))

  if (!is.null(timezone)) {
    # Side effect: check if time zone valid
    dbExecute(conn, paste0("SET time_zone = ", dbQuoteString(conn, timezone)))
  } else {
    timezone <- dbGetQuery(conn, "SELECT @@SESSION.time_zone")[[1]]
  }

  # Check if this is a valid time zone in R:
  timezone <- check_tz(timezone)

  if (is.null(timezone_out)) {
    timezone_out <- timezone
  } else {
    timezone_out <- check_tz(timezone_out)
  }

  conn@timezone <- timezone
  conn@timezone_out <- timezone_out

  dbExecute(conn, "SET autocommit = 0")
  on.exit(NULL)

  conn
}

#' @rdname dbConnect-MariaDBDriver-method
#' @export
setMethod("dbConnect", "MariaDBDriver", dbConnect_MariaDBDriver)

vlapply <- function(X, FUN, ..., USE.NAMES = TRUE) {
  vapply(X = X, FUN = FUN, FUN.VALUE = logical(1L), ..., USE.NAMES = USE.NAMES)
}

vcapply <- function(X, FUN, ..., USE.NAMES = TRUE) {
  vapply(X = X, FUN = FUN, FUN.VALUE = character(1L), ..., USE.NAMES = USE.NAMES)
}

stopc <- function(...) {
  stop(..., call. = FALSE, domain = NA)
}

warningc <- function(...) {
  warning(..., call. = FALSE, domain = NA)
}

sqlData_MariaDBConnection <- function(con, value, row.names = FALSE, ...) {
  value <- sql_data(value, con, row.names)
  value <- quote_string(value, con)

  value
}

setMethod("sqlData", "MariaDBConnection", sqlData_MariaDBConnection)

#' @rdname result-meta
#' @usage NULL
dbGetRowCount_MariaDBResult <- function(res, ...) {
  result_rows_fetched(res@ptr)
}

#' @rdname result-meta
#' @export
setMethod("dbGetRowCount", "MariaDBResult", dbGetRowCount_MariaDBResult)

#' @rdname result-meta
#' @usage NULL
dbGetRowsAffected_MariaDBResult <- function(res, ...) {
  result_rows_affected(res@ptr)
}

#' @rdname result-meta
#' @export
setMethod("dbGetRowsAffected", "MariaDBResult", dbGetRowsAffected_MariaDBResult)

#' @name mysql-quoting
#' @usage NULL
dbQuoteLiteral_MySQLConnection <- function(conn, x, ...) {
  # Switchpatching to avoid ambiguous S4 dispatch, so that our method
  # is used only if no alternatives are available.

  if (inherits(x, "POSIXt")) {
    return(dbQuoteString(
      conn,
      strftime(as.POSIXct(x), "%Y%m%d%H%M%S", tz = "UTC")
    ))
  }
  callNextMethod()
}

#' @rdname mysql-quoting
#' @export
setMethod("dbQuoteLiteral", signature("MySQLConnection"), dbQuoteLiteral_MySQLConnection)

#' @rdname query
#' @usage NULL
dbSendStatement_MariaDBConnection_character <- function(conn, statement, params = NULL, ..., immediate = FALSE) {
  dbSend(conn, statement, params, is_statement = TRUE, immediate)
}

#' @rdname query
#' @export
setMethod("dbSendStatement", signature("MariaDBConnection", "character"), dbSendStatement_MariaDBConnection_character)

# dbSendQuery()
# dbSendStatement()
# dbDataType()
# dbQuoteString()
# dbQuoteIdentifier()
# dbWriteTable()
# dbReadTable()
# dbListTables()
# dbExistsTable()
# dbListFields()
# dbRemoveTable()
# dbGetInfo()
#' @rdname MariaDBConnection-class
#' @usage NULL
dbGetInfo_MariaDBConnection <- function(dbObj, what = "", ...) {
  connection_info(dbObj@ptr)
}

#' @rdname MariaDBConnection-class
#' @export
setMethod("dbGetInfo", "MariaDBConnection", dbGetInfo_MariaDBConnection)

#' @rdname mariadb-tables
#' @param fail_if_missing If `FALSE`, `dbRemoveTable()` succeeds if the
#'   table doesn't exist.
#' @usage NULL
dbRemoveTable_MariaDBConnection_character <- function(conn, name, ..., temporary = FALSE, fail_if_missing = TRUE) {
  extra <- list(...)

  name <- dbQuoteIdentifier(conn, name)
  dbExecute(
    conn,
    paste0(
      "DROP ",
      if (temporary) "TEMPORARY ",
      "TABLE ",
      if (!fail_if_missing) "IF EXISTS ",
      name
    )
  )
  invisible(TRUE)
}

#' @rdname mariadb-tables
#' @export
setMethod("dbRemoveTable", c("MariaDBConnection", "character"), dbRemoveTable_MariaDBConnection_character)

#' @include MariaDBConnection.R
#' @include MariaDBResult.R
NULL

convert_bigint <- function(df, bigint) {
  if (bigint == "integer64") return(df)
  is_int64 <- which(vlapply(df, inherits, "integer64"))
  if (length(is_int64) == 0) return(df)

  as_bigint <- switch(bigint,
    integer = as.integer,
    numeric = as.numeric,
    character = as.character
  )

  df[is_int64] <- suppressWarnings(lapply(df[is_int64], as_bigint))
  df
}

fix_timezone <- function(ret, conn) {
  is_datetime <- which(vapply(ret, inherits, "POSIXt", FUN.VALUE = logical(1)))
  if (length(is_datetime) > 0) {
    ret[is_datetime] <- lapply(ret[is_datetime], function(x) {
      x <- lubridate::with_tz(x, "UTC")
      x <- lubridate::force_tz(x, conn@timezone)
      lubridate::with_tz(x, conn@timezone_out)
    })
  }

  ret
}

fix_blob <- function(ret) {
  is_blob <- which(vapply(ret, is.list, FUN.VALUE = logical(1)))
  if (length(is_blob) > 0) {
    ret[is_blob] <- lapply(ret[is_blob], blob::as_blob)
  }

  ret
}

dbSend <- function(conn, statement, params = NULL, is_statement, immediate) {
  statement <- enc2utf8(statement)

  rs <- new("MariaDBResult",
    sql = statement,
    ptr = result_create(conn@ptr, statement, is_statement, immediate),
    bigint = conn@bigint,
    conn = conn
  )

  on.exit(dbClearResult(rs))

  if (!is.null(params)) {
    dbBind(rs, params)
  }

  on.exit(NULL)
  rs
}

#' Database interface meta-data.
#'
#' See documentation of generics for more details.
#'
#' @param res An object of class [MariaDBResult-class]
#' @param ... Ignored. Needed for compatibility with generic
#' @examples
#' if (mariadbHasDefault()) {
#'   con <- dbConnect(RMariaDB::MariaDB(), dbname = "test")
#'   dbWriteTable(con, "t1", datasets::USArrests, temporary = TRUE)
#'
#'   rs <- dbSendQuery(con, "SELECT * FROM t1 WHERE UrbanPop >= 80")
#'   rs
#'
#'   dbGetStatement(rs)
#'   dbHasCompleted(rs)
#'   dbColumnInfo(rs)
#'
#'   dbFetch(rs)
#'   rs
#'
#'   dbClearResult(rs)
#'   dbDisconnect(con)
#' }
#' @name result-meta
NULL

#' @include MariaDBConnection.R
NULL

check_tz <- function(timezone) {
  arg_name <- deparse(substitute(timezone))

  if (timezone == "+00:00") {
    timezone <- "UTC"
  }

  tryCatch(
    lubridate::force_tz(as.POSIXct("2021-03-01 10:40"), timezone),
    error = function(e) {
      warning(
        "Invalid time zone '", timezone, "', ",
        "falling back to local time.\n",
        "Set the `", arg_name, "` argument to a valid time zone.\n",
        conditionMessage(e),
        call. = FALSE
      )
      timezone <- ""
    }
  )

  timezone
}

#' @export
#' @import methods DBI
#' @importFrom hms hms
#' @importFrom bit64 integer64 is.integer64
#' @rdname dbConnect-MariaDBDriver-method
#' @examples
#' if (mariadbHasDefault()) {
#'   # connect to a database and load some data
#'   con <- dbConnect(RMariaDB::MariaDB(), dbname = "test")
#'   dbWriteTable(con, "USArrests", datasets::USArrests, temporary = TRUE)
#'
#'   # query
#'   rs <- dbSendQuery(con, "SELECT * FROM USArrests")
#'   d1 <- dbFetch(rs, n = 10)      # extract data in chunks of 10 rows
#'   dbHasCompleted(rs)
#'   d2 <- dbFetch(rs, n = -1)      # extract all remaining data
#'   dbHasCompleted(rs)
#'   dbClearResult(rs)
#'   dbListTables(con)
#'
#'   # clean up
#'   dbDisconnect(con)
#' }
MariaDB <- function() {
  new("MariaDBDriver")
}

#' Client flags
#'
#' Use for the `client.flag` argument to [dbConnect()], multiple flags can be
#' combined with `+` or [bitwOr()].
#' The flags are provided for completeness.
#' To enforce SSL for the DB connection, add the flag `CLIENT_SSL`.
#'
#' @seealso
#' The `flags` argument at https://mariadb.com/kb/en/library/mysql_real_connect.
#'
#' @examples
#' \dontrun{
#' library(DBI)
#' library(RMariaDB)
#' con1 <- dbConnect(MariaDB(), client.flag = CLIENT_COMPRESS)
#' con2 <- dbConnect(
#'   MariaDB(),
#'   client.flag = bitwOr(CLIENT_COMPRESS, CLIENT_SSL)
#' )
#' }
#'
#' @aliases CLIENT_LONG_PASSWORD CLIENT_FOUND_ROWS CLIENT_LONG_FLAG
#' CLIENT_CONNECT_WITH_DB CLIENT_NO_SCHEMA CLIENT_COMPRESS CLIENT_ODBC
#' CLIENT_LOCAL_FILES CLIENT_IGNORE_SPACE CLIENT_PROTOCOL_41 CLIENT_INTERACTIVE
#' CLIENT_SSL CLIENT_IGNORE_SIGPIPE CLIENT_TRANSACTIONS CLIENT_RESERVED CLIENT_RESERVED2
#' CLIENT_MULTI_STATEMENTS CLIENT_MULTI_RESULTS CLIENT_SSL_VERIFY_SERVER_CERT
#' @name Client-flags
NULL

## The following client flags were copied from mysql_com.h (version 4.1.13)
## but it may not make sense to set some of this from RMariaDB.
#' @export
CLIENT_LONG_PASSWORD <-   1
# new more secure passwords

#' @export
CLIENT_FOUND_ROWS    <-   2
# Found instead of affected rows

#' @export
CLIENT_LONG_FLAG     <-   4
# Get all column flags

#' @export
CLIENT_CONNECT_WITH_DB <- 8
# One can specify db on connect

#' @export
CLIENT_NO_SCHEMA     <-  16
# Don't allow database.table.column

#' @export
CLIENT_COMPRESS      <-  32
# Can use compression protocol

#' @export
CLIENT_ODBC          <-  64
# Odbc client

#' @export
CLIENT_LOCAL_FILES   <- 128
# Can use LOAD DATA LOCAL

#' @export
CLIENT_IGNORE_SPACE  <- 256
# Ignore spaces before '('

#' @export
CLIENT_PROTOCOL_41   <- 512
# New 4.1 protocol

#' @export
CLIENT_INTERACTIVE   <- 1024
# This is an interactive client

#' @export
CLIENT_SSL           <- 2048
# Switch to SSL after handshake

#' @export
CLIENT_IGNORE_SIGPIPE <- 4096
# IGNORE sigpipes

#' @export
CLIENT_TRANSACTIONS <- 8192
# Client knows about transactions

#' @export
CLIENT_RESERVED     <- 16384
# Old flag for 4.1 protocol

#' @export
CLIENT_RESERVED2    <- 32768
# Old flag for 4.1 authentication

#' @export
CLIENT_MULTI_STATEMENTS  <- 65536
# Enable/disable multi-stmt support

#' @export
CLIENT_MULTI_RESULTS     <- 131072
# Enable/disable multi-results

#' @export
CLIENT_SSL_VERIFY_SERVER_CERT <- 1073741824
# Enable/disable verifying server SSL certificate

#' @rdname mariadb-quoting
#' @usage NULL
dbUnquoteIdentifier_MariaDBConnection_SQL <- function(conn, x, ...) {
  id_rx <- "(?:`((?:[^`]|``)+)`|([^`. ]+))"

  rx <- paste0(
    "^",
    "(?:|", id_rx, "[.])",
    "(?:|", id_rx, ")",
    "$"
  )

  bad <- grep(rx, x, invert = TRUE)
  if (length(bad) > 0) {
    stop("Can't unquote ", x[bad[[1]]], call. = FALSE)
  }
  schema <- gsub(rx, "\\1\\2", x)
  schema <- gsub("``", "`", schema)
  table <- gsub(rx, "\\3\\4", x)
  table <- gsub("``", "`", table)

  ret <- Map(schema, table, f = as_table)
  names(ret) <- names(x)
  ret
}

#' @rdname mariadb-quoting
#' @export
setMethod("dbUnquoteIdentifier", c("MariaDBConnection", "SQL"), dbUnquoteIdentifier_MariaDBConnection_SQL)

#' @rdname mariadb-tables
#' @usage NULL
dbReadTable_MariaDBConnection_character <- function(conn, name, ..., row.names = FALSE, check.names = TRUE) {
  row.names <- compatRowNames(row.names)

  if ((!is.logical(row.names) && !is.character(row.names)) || length(row.names) != 1L) {
    stopc("`row.names` must be a logical scalar or a string")
  }

  if (!is.logical(check.names) || length(check.names) != 1L) {
    stopc("`check.names` must be a logical scalar")
  }

  name <- dbQuoteIdentifier(conn, name)
  out <- dbGetQuery(conn, paste("SELECT * FROM ", name),
    row.names = row.names
  )

  if (check.names) {
    names(out) <- make.names(names(out), unique = TRUE)
  }

  out
}

#' @rdname mariadb-tables
#' @export
setMethod("dbReadTable", c("MariaDBConnection", "character"), dbReadTable_MariaDBConnection_character)

#' @rdname mariadb-quoting
#' @usage NULL
dbQuoteIdentifier_MariaDBConnection_SQL <- function(conn, x, ...) {
  x
}

#' @rdname mariadb-quoting
#' @export
setMethod("dbQuoteIdentifier", c("MariaDBConnection", "SQL"), dbQuoteIdentifier_MariaDBConnection_SQL)

#' @rdname mariadb-quoting
#' @usage NULL
dbQuoteIdentifier_MariaDBConnection_character <- function(conn, x, ...) {
  if (any(is.na(x))) {
    stop("Cannot pass NA to dbQuoteIdentifier()", call. = FALSE)
  }
  x <- gsub("`", "``", x, fixed = TRUE)
  if (length(x) == 0L) {
    SQL(character(), names = names(x))
  } else {
    # Not calling encodeString() here to keep things simple
    SQL(paste("`", x, "`", sep = ""), names = names(x))
  }
}

#' @rdname mariadb-quoting
#' @export
setMethod("dbQuoteIdentifier", c("MariaDBConnection", "character"), dbQuoteIdentifier_MariaDBConnection_character)

#' @rdname MariaDBResult-class
#' @usage NULL
dbIsValid_MariaDBResult <- function(dbObj, ...) {
  result_valid(dbObj@ptr)
}

#' @rdname MariaDBResult-class
#' @export
setMethod("dbIsValid", "MariaDBResult", dbIsValid_MariaDBResult)

#' @rdname MariaDBDriver-class
#' @usage NULL
dbIsValid_MariaDBDriver <- function(dbObj, ...) {
  TRUE
}

#' @rdname MariaDBDriver-class
#' @export
setMethod("dbIsValid", "MariaDBDriver", dbIsValid_MariaDBDriver)

#' @rdname mariadb-quoting
#' @usage NULL
dbQuoteString_MariaDBConnection_SQL <- function(conn, x, ...) {
  x
}

#' @rdname mariadb-quoting
#' @export
setMethod("dbQuoteString", c("MariaDBConnection", "SQL"), dbQuoteString_MariaDBConnection_SQL)

#' Check if default database is available.
#'
#' RMariaDB examples and tests connect to a database defined by the
#' `rs-dbi` group in `~/.my.cnf`. This function checks if that
#' database is available, and if not, displays an informative message.
#' `mariadbDefault()` works similarly but throws a testthat skip condition
#' on failure, making it suitable for use in tests.
#'
#' @export
#' @examples
#' if (mariadbHasDefault()) {
#'   db <- dbConnect(RMariaDB::MariaDB(), dbname = "test")
#'   dbListTables(db)
#'   dbDisconnect(db)
#' }
mariadbHasDefault <- function() {
  tryCatch(
    {
      mariadb_default()
      TRUE
    },
    error = function(...) {
      message(
        "Could not initialise default MariaDB database. If MariaDB is running\n",
        "check that you have a ~/.my.cnf file that contains a [rs-dbi] section\n",
        "describing how to connect to a test database."
      )
      FALSE
    })
}

#' @export
#' @rdname mariadbHasDefault
mariadbDefault <- function() {
  tryCatch(
    {
      mariadb_default()
    },
    error = function(...) {
      testthat::skip("Test database not available")
    }
  )
}

mysqlDefault <- function() {
  tryCatch(
    {
      mariadb_default(mysql = TRUE)
    },
    error = function(...) {
      testthat::skip("Test database not available")
    }
  )
}

mariadbForceDefault <- function() {
  tryCatch(
    {
      mariadb_default(mysql = FALSE)
    },
    error = function(...) {
      testthat::skip("Test database not available")
    }
  )
}

mariadb_default <- function(...) {
  rlang::inject(dbConnect(MariaDB(), !!!mariadb_default_args, ...))
}

mariadb_default_args <- as.list(c(
  dbname = "test",
  # host = "192.168.64.2",
  # user = "compose",
  # password = "YourStrong!Passw0rd",
  NULL
))

# test_that("can't override existing table with default options", {
#   con <- mariadbDefault()
#
#   x <- data.frame(col1 = 1:10, col2 = letters[1:10])
#   dbWriteTable(con, "t1", x, temporary = TRUE)
#   expect_error(dbWriteTable(con, "t1", x), "exists in database")
#   dbDisconnect(con)
# })

# Not generic enough for DBItest
test_that("dbWriteTable() throws error if constraint violated", {
  con <- mariadbDefault()
  on.exit(dbDisconnect(con))

  x <- data.frame(col1 = 1:10, col2 = letters[1:10])

  dbWriteTable(con, "t1", x[1:3, ], overwrite = TRUE, temporary = TRUE)
  dbExecute(con, "CREATE UNIQUE INDEX t1_c1_c2_idx ON t1(col1, col2(1))")
  expect_error(dbWriteTable(con, "t1", x, append = TRUE), "added 7 rows|Duplicate entry")
})

test_that("dbAppendTable() throws error if constraint violated", {
  con <- mariadbDefault()
  on.exit(dbDisconnect(con))

  x <- data.frame(col1 = 1:10, col2 = letters[1:10])

  dbWriteTable(con, "t1", x[1:3, ], overwrite = TRUE, temporary = TRUE)
  dbExecute(con, "CREATE UNIQUE INDEX t1_c1_c2_idx ON t1(col1, col2(1))")
  expect_error(dbAppendTable(con, "t1", x), "added 7 rows|Duplicate entry")
})

test_that("dbAppendTable() works with Id", {
  con <- mariadbDefault()
  on.exit(dbDisconnect(con))

  dbExecute(con, "CREATE TEMPORARY TABLE t1(n integer)")
  expect_equal(dbAppendTable(con, Id(table = "t1"), data.frame(n = 1:10)), 10)
})

# Available only in MariaDB
test_that("can read file from disk", {
  con <- mariadbDefault()
  on.exit(dbDisconnect(con))

  if (dbGetQuery(con, "SHOW VARIABLES LIKE 'local_infile'")$Value == "OFF") {
    skip("local_infile is set to OFF, can't test LOAD DATA INFILE")
  }

  expected <- data.frame(
    a = c(1:3, NA),
    b = c("x", "y", "z", "E"),
    stringsAsFactors = FALSE
  )

  dbWriteTable(con, "dat", "dat-n.bin", sep = "|", eol = "\n",
    temporary = TRUE, overwrite = TRUE)
  expect_equal(dbReadTable(con, "dat"), expected)
})

test_that("converts NaN and Inf to NULL", {
  con <- mariadbDefault()
  on.exit(dbDisconnect(con))

  x <- data.frame(col1 = c(-Inf, NA, 0, NaN, Inf))

  dbWriteTable(con, "t1", x, overwrite = TRUE, temporary = TRUE)
  expect_equal(dbReadTable(con, "t1"), data.frame(col1 = c(NA, NA, 0, NA, NA)))
})

test_that("write dates prior to 1970 (#232)", {
  con <- mariadbDefault()
  on.exit(dbDisconnect(con))

  x <- data.frame(col1 = as.Date("1970-01-01") + 2:-2)

  dbWriteTable(con, "t1", x, overwrite = TRUE, temporary = TRUE)
  expect_equal(dbReadTable(con, "t1"), x)
})

test_that("writing and reading JSON (#127)", {
  con <- mariadbDefault()
  on.exit(dbDisconnect(con))

  x <- data.frame(col1 = "[1,2,3]", stringsAsFactors = FALSE)

  dbWriteTable(con, "t1", x, field.types = c(col1 = "json"), overwrite = TRUE, temporary = TRUE)

  suppressWarnings(d <- dbReadTable(con, "t1"))

  # MySQL 8 returns "[1, 2, 3]", while MariaDB returns "[1,2,3]"
  d$col1 <- gsub('\\s', '', d$col1)

  expect_equal(d, x)
})

test_that("quoting string", {
  con <- mariadbDefault()
  on.exit(dbDisconnect(con))

  quoted <- dbQuoteIdentifier(con, "Robert'); DROP TABLE Students;--")
  expect_s4_class(quoted, 'SQL')
  expect_equal(as.character(quoted),
    "`Robert'); DROP TABLE Students;--`")
})

test_that("quoting SQL", {
  con <- mariadbDefault()
  on.exit(dbDisconnect(con))

  quoted <- dbQuoteIdentifier(con, SQL("Robert'); DROP TABLE Students;--"))
  expect_s4_class(quoted, 'SQL')
  expect_equal(as.character(quoted),
    "Robert'); DROP TABLE Students;--")
})

test_that("quoting Id", {
  con <- mariadbDefault()
  on.exit(dbDisconnect(con))

  quoted <- dbQuoteIdentifier(con, Id(schema = 'Robert', table = 'Students;--'))
  expect_s4_class(quoted, 'SQL')
  expect_equal(as.character(quoted),
    "`Robert`.`Students;--`")
})

test_that("quoting Id with column, #254", {
  con <- mariadbDefault()
  on.exit(dbDisconnect(con))

  quoted <- dbQuoteIdentifier(con, Id(schema = 'Robert', table = 'Students;--', column = "dormitory"))
  expect_s4_class(quoted, 'SQL')
  expect_equal(as.character(quoted),
    "`Robert`.`Students;--`.`dormitory`")
})

test_that("quoting Id with column, unordered", {
  con <- mariadbDefault()
  on.exit(dbDisconnect(con))

  quoted <- dbQuoteIdentifier(con, Id(column = "dormitory", table = 'Students;--'))
  expect_s4_class(quoted, 'SQL')
  expect_equal(as.character(quoted),
    "`Students;--`.`dormitory`")
})

test_that("quoting errors", {
  con <- mariadbDefault()
  on.exit(dbDisconnect(con))

  expect_error(Id(table = 'Robert', table = 'Students;--'))
})

test_that("unquoting identifier - SQL with quotes", {
  con <- mariadbDefault()
  on.exit(dbDisconnect(con))

  expect_equal(dbUnquoteIdentifier(con, SQL('`Students;--`')),
    list(Id(table = 'Students;--')))

  expect_equal(dbUnquoteIdentifier(con, SQL('`Robert`.`Students;--`')),
    list(Id(schema = 'Robert', table = 'Students;--')))

  expect_equal(dbUnquoteIdentifier(con, SQL('`Rob``ert`.`Students;--`')),
    list(Id(schema = 'Rob`ert', table = 'Students;--')))

  expect_equal(dbUnquoteIdentifier(con, SQL('`Rob.ert`.`Students;--`')),
    list(Id(schema = 'Rob.ert',  table = 'Students;--')))

  expect_error(dbUnquoteIdentifier(con, SQL('`Robert.`Students`')),
    "^Can't unquote")
})

test_that("unquoting identifier - SQL without quotes", {
  con <- mariadbDefault()
  on.exit(dbDisconnect(con))

  expect_equal(dbUnquoteIdentifier(con, SQL('Students')),
    list(Id(table = 'Students')))

  expect_equal(dbUnquoteIdentifier(con, SQL('Robert.Students')),
    list(Id(schema = 'Robert', table = 'Students')))

  expect_error(dbUnquoteIdentifier(con, SQL('Rob``ert.Students')),
    "^Can't unquote")
})

test_that("unquoting identifier - Id", {
  con <- mariadbDefault()
  on.exit(dbDisconnect(con))

  expect_equal(dbUnquoteIdentifier(con,
    Id(schema = 'Robert', table = 'Students;--')),
  list(Id(schema = 'Robert', table = 'Students;--')))
})

# Can't test this in a generic fashion
test_that("setting parameter query is always complete", {
  conn <- mariadbDefault()
  rs <- dbSendStatement(conn, 'SET time_zone = "+00:00"')

  expect_true(dbHasCompleted(rs))

  dbClearResult(rs)
  dbDisconnect(conn)
})

# Maybe also relevant for other backends? move to DBItest?
test_that("prepared statements and SHOW queries", {
  conn <- mariadbDefault()
  rs <- dbGetQuery(conn, "SHOW PLUGINS")

  expect_gte(nrow(rs), 1L)

  dbDisconnect(conn)
})

test_that("fractional seconds in datetime (#170)", {
  con <- mariadbDefault()

  dataframe <- data.frame(DateTime = Sys.time())

  dbWriteTable(
    con,
    "my_table",
    dataframe,
    overwrite = TRUE,
    temporary = TRUE,
    field.types = c(DateTime = "datetime(6)"),
    row.names = FALSE
  )

  out <- dbReadTable(con, "my_table")
  expect_equal(round(as.numeric(out$DateTime) * 1e6), round(as.numeric(dataframe$DateTime) * 1e6))

  out <- dbGetQuery(con, "SELECT ADDTIME(NOW(), 0.5) AS DateTime")
  seconds <- as.numeric(out$DateTime)
  expect_equal(seconds - floor(seconds), 0.5)

  dbDisconnect(con)
})

test_that("fractional seconds in time (#288)", {
  con <- mariadbDefault()

  time_str <- "09:30:01.123456"
  time_hms <- hms::as_hms(time_str)
  dataframe <- data.frame(Time = time_hms)

  dbWriteTable(
    con,
    "my_table",
    dataframe,
    overwrite = TRUE,
    temporary = TRUE,
    field.types = c(Time = "time(6)"),
    row.names = FALSE
  )

  out <- dbReadTable(con, "my_table")
  expect_equal(dataframe, out)
  expect_equal(out$Time, time_hms)
  expect_equal(format(out$Time), time_str)

  dbDisconnect(con)
})

test_that("timezone argument (#184)", {
  skip_on_cran()

  expect_warning(
    conn <- mariadb_default(timezone = "+02:00"),
    "Invalid time zone"
  )
  tz <- dbGetQuery(conn, "SELECT @@session.time_zone")
  expect_equal(tz[[1]], "+02:00")
  dbDisconnect(conn)
})

test_that("bit(1) fields support NA values (#201)", {
  con <- mariadbDefault()
  on.exit(dbDisconnect(con))

  dbWriteTable(con, "testbit1",
    data.frame(a = c(NA_integer_, 0:1), b = 1:3),
    field.types = c(a = "BIT(1)"), overwrite = TRUE
  )
  result <- dbGetQuery(con, "SELECT a FROM testbit1 ORDER BY b")$a
  expect_equal(result, c(NA, FALSE, TRUE))

  dbRemoveTable(con, "testbit1")
})

test_that("Connecting with mysql = FALSE", {
  con <- mariadbForceDefault()
  on.exit(dbDisconnect(con))

  expect_s4_class(con, "MariaDBConnection")
  expect_false(is(con, "MySQLConnection"))
})

test_that("Connecting with mysql = TRUE", {
  con <- mysqlDefault()
  on.exit(dbDisconnect(con))

  expect_s4_class(con, "MariaDBConnection")
  expect_s4_class(con, "MySQLConnection")
})

test_that("Connecting with mysql unset", {
  con <- mariadbDefault()
  on.exit(dbDisconnect(con))

  expect_s4_class(con, "MariaDBConnection")
})

test_that("dbQuoteIdentifier() and def", {
  skip_if_not(mariadbHasDefault())

  conn <- withr::local_db_connection(mariadbDefault())

  expect_equal(dbQuoteIdentifier(conn, Id("def", "a", "b")), SQL("`a`.`b`"))
  expect_equal(dbQuoteIdentifier(conn, Id("a", "b")), SQL("`a`.`b`"))
  expect_equal(dbQuoteIdentifier(conn, Id("def", "a")), SQL("`def`.`a`"))
})

skip_on_cran()

test_that("timestamp without time zone is returned correctly for TZ set (#190)", {
  withr::local_timezone("America/Chicago")

  query <- "SELECT CAST('2018-01-01 12:30:00' AS DATETIME) AS a, @@SESSION.time_zone AS b"

  db <- mariadb_default(timezone = NULL)
  res <- dbGetQuery(db, query)
  expect_equal(res[[1]], as.POSIXct("2018-01-01 12:30:00", tz = res[[2]]))
  dbDisconnect(db)

  db <- mariadb_default(timezone = "UTC")
  expect_equal(
    dbGetQuery(db, query)[[1]],
    as.POSIXct("2018-01-01 12:30:00", tz = "UTC")
  )
  dbDisconnect(db)

  db <- mariadb_default(timezone = "America/Chicago")
  expect_equal(
    dbGetQuery(db, query)[[1]],
    as.POSIXct("2018-01-01 12:30:00", tz = "America/Chicago")
  )
  dbDisconnect(db)

  db <- mariadb_default(timezone = "America/New_York")
  expect_equal(
    dbGetQuery(db, query)[[1]],
    as.POSIXct("2018-01-01 12:30:00", tz = "America/New_York")
  )
  dbDisconnect(db)

  db <- mariadb_default(timezone = "Europe/London")
  expect_equal(
    dbGetQuery(db, query)[[1]],
    as.POSIXct("2018-01-01 12:30:00", tz = "Europe/London")
  )
  dbDisconnect(db)

  db <- mariadb_default(timezone = "Europe/Zurich")
  expect_equal(
    dbGetQuery(db, query)[[1]],
    as.POSIXct("2018-01-01 12:30:00", tz = "Europe/Zurich")
  )
  dbDisconnect(db)
})

test_that("timestamp with time zone is returned correctly (#205)", {
  con <- mariadb_default()
  on.exit(dbDisconnect(con))

  withr::local_timezone("America/New_York")

  con <- mariadb_default(timezone = "America/Chicago")

  query <- "CREATE TEMPORARY TABLE junk AS
            (SELECT CAST('2020-05-04' AS DATETIME) AS ts)"
  res <- dbExecute(con, query)

  res <- dbGetQuery(con, "SELECT * FROM junk")
  expect_equal(res[[1]], as.POSIXct("2020-05-04", tz = "America/Chicago"))
})

test_that("timestamp without time zone is returned correctly (#221)", {
  con <- mariadb_default()
  on.exit(dbDisconnect(con))

  out <- dbGetQuery(con, "SELECT CAST('1960-01-01 12:00:00' AS DATETIME) AS before_epoch")
  expect_equal(as.Date(out[[1]]), as.Date("1960-01-01"))
})

test_that("timezone is passed on to the connection (#229)", {
  my_tz <- "US/Alaska"

  con <- mariadb_default(timezone = my_tz)
  on.exit(dbDisconnect(con))

  example <- data.frame(val = 0:71)
  example$ts <-
    lubridate::ymd_hms("2019-04-30 00:00:00", tz = my_tz) +
    lubridate::hours(example$val)

  dbWriteTable(
    con, "example", example, temporary = TRUE,
    overwrite = TRUE, append = FALSE
  )
  res <- dbReadTable(con, "example")
  expect_equal(res, example)

  query <- '
    SELECT date(ts) AS dte, MIN(val) AS min_val, MAX(val) AS max_val
    FROM example
    GROUP BY dte
    ORDER BY dte'
  expected <- data.frame(
    dte = as.Date("2019-04-30") + 0:2,
    min_val = 0:2 * 24,
    max_val = 0:2 * 24 + 23
  )

  res <- dbGetQuery(con, query)
  expect_equal(res, expected)
})

test_that("warning if time zone not interpretable", {
  skip_on_cran()

  expect_warning(con <- mariadb_default(timezone = "+01:00"))
  dbDisconnect(con)
  expect_warning(con <- mariadb_default(timezone_out = "+01:00"))
  dbDisconnect(con)
})

#' @rdname query
#' @usage NULL
dbClearResult_MariaDBResult <- function(res, ...) {
  if (!dbIsValid(res)) {
    warningc("Expired, result set already closed")
    return(invisible(TRUE))
  }
  result_release(res@ptr)
  invisible(TRUE)
}

#' @rdname query
#' @export
setMethod("dbClearResult", "MariaDBResult", dbClearResult_MariaDBResult)

#' @rdname mariadb-tables
#' @usage NULL
dbExistsTable_MariaDBConnection_character <- function(conn, name, ...) {
  stopifnot(length(name) == 1L)
  if (!dbIsValid(conn)) {
    stopc("Invalid connection")
  }
  tryCatch(
    {
      dbGetQuery(conn, paste0(
        "SELECT NULL FROM ", dbQuoteIdentifier(conn, name), " WHERE FALSE"
      ))
      TRUE
    },
    error = function(...) {
      FALSE
    }
  )
}

#' @rdname mariadb-tables
#' @export
setMethod("dbExistsTable", c("MariaDBConnection", "character"), dbExistsTable_MariaDBConnection_character)

#' @name result-meta
#' @usage NULL
dbColumnInfo_MariaDBResult <- function(res, ...) {
  df <- result_column_info(res@ptr)
  df
}

#' @rdname result-meta
#' @export
setMethod("dbColumnInfo", "MariaDBResult", dbColumnInfo_MariaDBResult)

# Generated by cpp11: do not edit by hand

connection_create <- function(host, user, password, db, port, unix_socket, client_flag, group, default_file, ssl_key, ssl_cert, ssl_ca, ssl_capath, ssl_cipher, timeout, reconnect) {
  .Call(`_RMariaDB_connection_create`, host, user, password, db, port, unix_socket, client_flag, group, default_file, ssl_key, ssl_cert, ssl_ca, ssl_capath, ssl_cipher, timeout, reconnect)
}

connection_valid <- function(con_) {
  .Call(`_RMariaDB_connection_valid`, con_)
}

connection_release <- function(con_) {
  invisible(.Call(`_RMariaDB_connection_release`, con_))
}

connection_info <- function(con) {
  .Call(`_RMariaDB_connection_info`, con)
}

connection_quote_string <- function(con, xs) {
  .Call(`_RMariaDB_connection_quote_string`, con, xs)
}

connection_begin_transaction <- function(con) {
  invisible(.Call(`_RMariaDB_connection_begin_transaction`, con))
}

connection_commit <- function(con) {
  invisible(.Call(`_RMariaDB_connection_commit`, con))
}

connection_rollback <- function(con) {
  invisible(.Call(`_RMariaDB_connection_rollback`, con))
}

connection_is_transacting <- function(con) {
  .Call(`_RMariaDB_connection_is_transacting`, con)
}

driver_init <- function() {
  invisible(.Call(`_RMariaDB_driver_init`))
}

driver_done <- function() {
  invisible(.Call(`_RMariaDB_driver_done`))
}

version <- function() {
  .Call(`_RMariaDB_version`)
}

result_create <- function(con, sql, is_statement, immediate) {
  .Call(`_RMariaDB_result_create`, con, sql, is_statement, immediate)
}

result_release <- function(res) {
  invisible(.Call(`_RMariaDB_result_release`, res))
}

result_valid <- function(res_) {
  .Call(`_RMariaDB_result_valid`, res_)
}

result_fetch <- function(res, n) {
  .Call(`_RMariaDB_result_fetch`, res, n)
}

result_bind <- function(res, params) {
  invisible(.Call(`_RMariaDB_result_bind`, res, params))
}

result_has_completed <- function(res) {
  .Call(`_RMariaDB_result_has_completed`, res)
}

result_rows_fetched <- function(res) {
  .Call(`_RMariaDB_result_rows_fetched`, res)
}

result_rows_affected <- function(res) {
  .Call(`_RMariaDB_result_rows_affected`, res)
}

result_column_info <- function(res) {
  .Call(`_RMariaDB_result_column_info`, res)
}

sql_data <- function(value, conn, row.names = FALSE, warn = FALSE) {
  row.names <- compatRowNames(row.names)
  value <- sqlRownamesToColumn(value, row.names)

  value <- factor_to_string(value, warn = warn)
  value <- string_to_utf8(value)
  value <- difftime_to_hms(value)
  value <- posixlt_to_posixct(value, conn@timezone)
  value <- numeric_to_finite(value)
  value <- date_to_double(value)
  value
}

factor_to_string <- function(value, warn = FALSE) {
  is_factor <- vlapply(value, is.factor)
  if (warn && any(is_factor)) {
    warning("Factors converted to character", call. = FALSE)
  }
  value[is_factor] <- lapply(value[is_factor], as.character)
  value
}

quote_string <- function(value, conn) {
  is_character <- vlapply(value, is.character)
  value[is_character] <- lapply(value[is_character], dbQuoteString, conn = conn)
  value
}

string_to_utf8 <- function(value) {
  is_char <- vlapply(value, is.character)
  value[is_char] <- lapply(value[is_char], enc2utf8)
  value
}

difftime_to_hms <- function(value) {
  is_difftime <- vlapply(value, inherits, "difftime")
  # https://github.com/tidyverse/hms/issues/84
  value[is_difftime] <- lapply(value[is_difftime], function(x) {
    mode(x) <- "double"
    hms::as_hms(x)
  })
  value
}

posixlt_to_posixct <- function(value, timezone) {
  is_posixlt <- vlapply(value, inherits, "POSIXt")
  value[is_posixlt] <- lapply(value[is_posixlt], function(x) {
    x <- as.POSIXct(x)

    # The database expects times as local time
    if (timezone != "UTC") {
      x <- lubridate::with_tz(x, timezone)
      x <- lubridate::force_tz(x, "UTC")
    }

    x
  })
  value
}

numeric_to_finite <- function(value) {
  is_numeric <- vlapply(value, is.numeric) & !vlapply(value, is.integer)
  value[is_numeric] <- lapply(value[is_numeric], function(x) {
    x[!is.finite(x)] <- NA
    x
  })
  value
}

date_to_double <- function(value) {
  is_date <- vlapply(value, inherits, "Date")
  value[is_date] <- lapply(value[is_date], function(x) {
    mode(x) <- "double"
    x
  })
  value
}

#' Class MariaDBResult
#'
#' MariaDB's query results class.  This classes encapsulates the result of an SQL
#' query or statement.
#'
#' @export
#' @keywords internal
setClass("MariaDBResult",
  contains = "DBIResult",
  slots = list(
    ptr = "externalptr",
    sql = "character",
    bigint = "character",
    conn = "MariaDBConnection"
  )
)

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dbQuoteLiteral_MySQLConnection.R,
%   R/quote-mysql.R
\name{mysql-quoting}
\alias{mysql-quoting}
\alias{dbQuoteLiteral_MySQLConnection}
\alias{dbQuoteLiteral,MySQLConnection-method}
\title{Quote MySQL strings and identifiers.}
\usage{
\S4method{dbQuoteLiteral}{MySQLConnection}(conn, x, ...)
}
\description{
In MySQL, identifiers are enclosed in backticks, e.g. \code{`x`}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MariaDBDriver.R
\name{mariadbClientLibraryVersions}
\alias{mariadbClientLibraryVersions}
\title{MariaDB Check for Compiled Versus Loaded Client Library Versions}
\usage{
mariadbClientLibraryVersions()
}
\value{
A named integer vector of length two, the first element
representing the compiled library version and the second element
representing the loaded client library version.
}
\description{
This function prints out the compiled and loaded client library versions.
}
\examples{
mariadbClientLibraryVersions()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dbBind_MariaDBResult.R,
%   R/dbClearResult_MariaDBResult.R, R/dbFetch_MariaDBResult.R,
%   R/dbGetStatement_MariaDBResult.R,
%   R/dbSendQuery_MariaDBConnection_character.R,
%   R/dbSendStatement_MariaDBConnection_character.R
\name{query}
\alias{query}
\alias{dbBind_MariaDBResult}
\alias{dbBind,MariaDBResult-method}
\alias{dbClearResult_MariaDBResult}
\alias{dbClearResult,MariaDBResult-method}
\alias{dbFetch_MariaDBResult}
\alias{dbFetch,MariaDBResult-method}
\alias{dbGetStatement_MariaDBResult}
\alias{dbGetStatement,MariaDBResult-method}
\alias{dbSendQuery_MariaDBConnection_character}
\alias{dbSendQuery,MariaDBConnection,character-method}
\alias{dbSendStatement_MariaDBConnection_character}
\alias{dbSendStatement,MariaDBConnection,character-method}
\title{Execute a SQL statement on a database connection.}
\usage{
\S4method{dbBind}{MariaDBResult}(res, params, ...)

\S4method{dbClearResult}{MariaDBResult}(res, ...)

\S4method{dbFetch}{MariaDBResult}(res, n = -1, ..., row.names = FALSE)

\S4method{dbGetStatement}{MariaDBResult}(res, ...)

\S4method{dbSendQuery}{MariaDBConnection,character}(conn, statement, params = NULL, ..., immediate = FALSE)

\S4method{dbSendStatement}{MariaDBConnection,character}(conn, statement, params = NULL, ..., immediate = FALSE)
}
\arguments{
\item{res}{A \linkS4class{MariaDBResult} object.}

\item{params}{A list of query parameters to be substituted into
a parameterised query.}

\item{...}{Unused. Needed for compatibility with generic.}

\item{n}{Number of rows to retrieve. Use -1 to retrieve all rows.}

\item{row.names}{Either \code{TRUE}, \code{FALSE}, \code{NA} or a string.

If \code{TRUE}, always translate row names to a column called "row_names".
If \code{FALSE}, never translate row names. If \code{NA}, translate
rownames only if they're a character vector.

A string is equivalent to \code{TRUE}, but allows you to override the
default name.

For backward compatibility, \code{NULL} is equivalent to \code{FALSE}.}

\item{conn}{A \linkS4class{MariaDBConnection} object.}

\item{statement}{A character vector of length one specifying the SQL
statement that should be executed.  Only a single SQL statement should be
provided.}

\item{immediate}{If TRUE, uses the \code{mysql_real_query()} API
instead of \code{mysql_stmt_init()}.
This allows passing multiple statements (with \link{CLIENT_MULTI_STATEMENTS})
and turns off the ability to pass parameters.}
}
\description{
To retrieve results a chunk at a time, use \code{\link[DBI:dbSendQuery]{DBI::dbSendQuery()}},
\code{\link[DBI:dbFetch]{DBI::dbFetch()}}, then \code{\link[DBI:dbClearResult]{DBI::dbClearResult()}}. Alternatively, if you want all the
results (and they'll fit in memory) use \code{\link[DBI:dbGetQuery]{DBI::dbGetQuery()}} which sends,
fetches and clears for you. For data manipulation queries (i.e. queries
that do not return data, such as \code{UPDATE}, \code{DELETE}, etc.),
\code{\link[DBI:dbSendStatement]{DBI::dbSendStatement()}} serves as a counterpart to \code{\link[DBI:dbSendQuery]{DBI::dbSendQuery()}}, while
\code{\link[DBI:dbExecute]{DBI::dbExecute()}} corresponds to \code{\link[DBI:dbGetQuery]{DBI::dbGetQuery()}}.
}
\examples{
if (mariadbHasDefault()) {
  con <- dbConnect(RMariaDB::MariaDB(), dbname = "test")
  dbWriteTable(con, "arrests", datasets::USArrests, temporary = TRUE)

  # Run query to get results as dataframe
  dbGetQuery(con, "SELECT * FROM arrests limit 3")

  # Send query to pull requests in batches
  res <- dbSendQuery(con, "SELECT * FROM arrests")
  data <- dbFetch(res, n = 2)
  data
  dbHasCompleted(res)

  dbClearResult(res)
  dbDisconnect(con)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dbColumnInfo_MariaDBResult.R,
%   R/dbGetRowCount_MariaDBResult.R, R/dbGetRowsAffected_MariaDBResult.R,
%   R/dbHasCompleted_MariaDBResult.R, R/query.R
\name{result-meta}
\alias{result-meta}
\alias{dbColumnInfo_MariaDBResult}
\alias{dbColumnInfo,MariaDBResult-method}
\alias{dbGetRowCount_MariaDBResult}
\alias{dbGetRowCount,MariaDBResult-method}
\alias{dbGetRowsAffected_MariaDBResult}
\alias{dbGetRowsAffected,MariaDBResult-method}
\alias{dbHasCompleted_MariaDBResult}
\alias{dbHasCompleted,MariaDBResult-method}
\title{Database interface meta-data.}
\usage{
\S4method{dbColumnInfo}{MariaDBResult}(res, ...)

\S4method{dbGetRowCount}{MariaDBResult}(res, ...)

\S4method{dbGetRowsAffected}{MariaDBResult}(res, ...)

\S4method{dbHasCompleted}{MariaDBResult}(res, ...)
}
\arguments{
\item{res}{An object of class \linkS4class{MariaDBResult}}

\item{...}{Ignored. Needed for compatibility with generic}
}
\description{
See documentation of generics for more details.
}
\examples{
if (mariadbHasDefault()) {
  con <- dbConnect(RMariaDB::MariaDB(), dbname = "test")
  dbWriteTable(con, "t1", datasets::USArrests, temporary = TRUE)

  rs <- dbSendQuery(con, "SELECT * FROM t1 WHERE UrbanPop >= 80")
  rs

  dbGetStatement(rs)
  dbHasCompleted(rs)
  dbColumnInfo(rs)

  dbFetch(rs)
  rs

  dbClearResult(rs)
  dbDisconnect(con)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dbQuoteIdentifier_MariaDBConnection_Id.R,
%   R/dbQuoteIdentifier_MariaDBConnection_SQL.R,
%   R/dbQuoteIdentifier_MariaDBConnection_character.R,
%   R/dbQuoteLiteral_MariaDBConnection.R,
%   R/dbQuoteString_MariaDBConnection_SQL.R,
%   R/dbQuoteString_MariaDBConnection_character.R,
%   R/dbUnquoteIdentifier_MariaDBConnection_SQL.R, R/quote.R
\name{mariadb-quoting}
\alias{mariadb-quoting}
\alias{dbQuoteIdentifier_MariaDBConnection_Id}
\alias{dbQuoteIdentifier,MariaDBConnection,Id-method}
\alias{dbQuoteIdentifier_MariaDBConnection_SQL}
\alias{dbQuoteIdentifier,MariaDBConnection,SQL-method}
\alias{dbQuoteIdentifier_MariaDBConnection_character}
\alias{dbQuoteIdentifier,MariaDBConnection,character-method}
\alias{dbQuoteLiteral_MariaDBConnection}
\alias{dbQuoteLiteral,MariaDBConnection-method}
\alias{dbQuoteString_MariaDBConnection_SQL}
\alias{dbQuoteString,MariaDBConnection,SQL-method}
\alias{dbQuoteString_MariaDBConnection_character}
\alias{dbQuoteString,MariaDBConnection,character-method}
\alias{dbUnquoteIdentifier_MariaDBConnection_SQL}
\alias{dbUnquoteIdentifier,MariaDBConnection,SQL-method}
\title{Quote MariaDB strings and identifiers.}
\usage{
\S4method{dbQuoteIdentifier}{MariaDBConnection,Id}(conn, x, ...)

\S4method{dbQuoteIdentifier}{MariaDBConnection,SQL}(conn, x, ...)

\S4method{dbQuoteIdentifier}{MariaDBConnection,character}(conn, x, ...)

\S4method{dbQuoteLiteral}{MariaDBConnection}(conn, x, ...)

\S4method{dbQuoteString}{MariaDBConnection,SQL}(conn, x, ...)

\S4method{dbQuoteString}{MariaDBConnection,character}(conn, x, ...)

\S4method{dbUnquoteIdentifier}{MariaDBConnection,SQL}(conn, x, ...)
}
\description{
In MariaDB, identifiers are enclosed in backticks, e.g. \code{`x`}.
}
\examples{
if (mariadbHasDefault()) {
  con <- dbConnect(RMariaDB::MariaDB())
  dbQuoteIdentifier(con, c("a b", "a`b"))
  dbQuoteString(con, c("a b", "a'b"))
  dbDisconnect(con)
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/default.R
\name{mariadbHasDefault}
\alias{mariadbHasDefault}
\alias{mariadbDefault}
\title{Check if default database is available.}
\usage{
mariadbHasDefault()

mariadbDefault()
}
\description{
RMariaDB examples and tests connect to a database defined by the
\code{rs-dbi} group in \verb{~/.my.cnf}. This function checks if that
database is available, and if not, displays an informative message.
\code{mariadbDefault()} works similarly but throws a testthat skip condition
on failure, making it suitable for use in tests.
}
\examples{
if (mariadbHasDefault()) {
  db <- dbConnect(RMariaDB::MariaDB(), dbname = "test")
  dbListTables(db)
  dbDisconnect(db)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dbDataType_MariaDBConnection.R,
%   R/dbDataType_MariaDBDriver.R
\name{dbDataType}
\alias{dbDataType}
\alias{dbDataType_MariaDBConnection}
\alias{dbDataType,MariaDBConnection-method}
\alias{dbDataType_MariaDBDriver}
\alias{dbDataType,MariaDBDriver-method}
\title{Determine the SQL Data Type of an S object}
\usage{
\S4method{dbDataType}{MariaDBConnection}(dbObj, obj, ...)

\S4method{dbDataType}{MariaDBDriver}(dbObj, obj, ...)
}
\arguments{
\item{dbObj}{A \linkS4class{MariaDBDriver} or \linkS4class{MariaDBConnection} object.}

\item{obj}{R/S-Plus object whose SQL type we want to determine.}

\item{\dots}{any other parameters that individual methods may need.}
}
\description{
This method is a straight-forward implementation of the corresponding
generic function.
}
\examples{
dbDataType(RMariaDB::MariaDB(), "a")
dbDataType(RMariaDB::MariaDB(), 1:3)
dbDataType(RMariaDB::MariaDB(), 2.5)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/connect.R
\name{Client-flags}
\alias{Client-flags}
\alias{CLIENT_LONG_PASSWORD}
\alias{CLIENT_FOUND_ROWS}
\alias{CLIENT_LONG_FLAG}
\alias{CLIENT_CONNECT_WITH_DB}
\alias{CLIENT_NO_SCHEMA}
\alias{CLIENT_COMPRESS}
\alias{CLIENT_ODBC}
\alias{CLIENT_LOCAL_FILES}
\alias{CLIENT_IGNORE_SPACE}
\alias{CLIENT_PROTOCOL_41}
\alias{CLIENT_INTERACTIVE}
\alias{CLIENT_SSL}
\alias{CLIENT_IGNORE_SIGPIPE}
\alias{CLIENT_TRANSACTIONS}
\alias{CLIENT_RESERVED}
\alias{CLIENT_RESERVED2}
\alias{CLIENT_MULTI_STATEMENTS}
\alias{CLIENT_MULTI_RESULTS}
\alias{CLIENT_SSL_VERIFY_SERVER_CERT}
\title{Client flags}
\description{
Use for the \code{client.flag} argument to \code{\link[DBI:dbConnect]{DBI::dbConnect()}}, multiple flags can be
combined with \code{+} or \code{\link[=bitwOr]{bitwOr()}}.
The flags are provided for completeness.
To enforce SSL for the DB connection, add the flag \code{CLIENT_SSL}.
}
\examples{
\dontrun{
library(DBI)
library(RMariaDB)
con1 <- dbConnect(MariaDB(), client.flag = CLIENT_COMPRESS)
con2 <- dbConnect(
  MariaDB(),
  client.flag = bitwOr(CLIENT_COMPRESS, CLIENT_SSL)
)
}

}
\seealso{
The \code{flags} argument at https://mariadb.com/kb/en/library/mysql_real_connect.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MariaDBConnection.R,
%   R/dbDisconnect_MariaDBConnection.R, R/dbGetInfo_MariaDBConnection.R,
%   R/dbIsValid_MariaDBConnection.R, R/show_MariaDBConnection.R
\docType{class}
\name{MariaDBConnection-class}
\alias{MariaDBConnection-class}
\alias{MariaDBConnection}
\alias{MySQLConnection}
\alias{MySQLConnection-class}
\alias{format.MariaDBConnection}
\alias{dbDisconnect_MariaDBConnection}
\alias{dbDisconnect,MariaDBConnection-method}
\alias{dbGetInfo_MariaDBConnection}
\alias{dbGetInfo,MariaDBConnection-method}
\alias{dbIsValid_MariaDBConnection}
\alias{dbIsValid,MariaDBConnection-method}
\alias{show_MariaDBConnection}
\alias{show,MariaDBConnection-method}
\title{Class MariaDBConnection.}
\usage{
\method{format}{MariaDBConnection}(x, ...)

\S4method{dbDisconnect}{MariaDBConnection}(conn, ...)

\S4method{dbGetInfo}{MariaDBConnection}(dbObj, what = "", ...)

\S4method{dbIsValid}{MariaDBConnection}(dbObj, ...)

\S4method{show}{MariaDBConnection}(object)
}
\description{
\code{"MariaDBConnection"} objects are usually created by \code{\link[DBI:dbConnect]{DBI::dbConnect()}}.
They represent a connection to a MariaDB or MySQL database.
}
\details{
The \code{"MySQLConnection"} class is a subclass of \code{"MariaDBConnection"}.
Objects of that class are created by \code{dbConnect(MariaDB(), ..., mysql = TRUE)}
to indicate that the server is a MySQL server.
The \pkg{RMariaDB} package supports both MariaDB and MySQL servers, but the SQL dialect
and other details vary.
The default is to detect the server type based on the version number.

The older \pkg{RMySQL} package also implements the \code{"MySQLConnection"} class.
The S4 system is able to distinguish between \pkg{RMariaDB} and \pkg{RMySQL} objects
even if both packages are loaded.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RMariaDB-package.R
\docType{package}
\name{RMariaDB-package}
\alias{RMariaDB}
\alias{RMariaDB-package}
\title{RMariaDB: Database Interface and MariaDB Driver}
\description{
Implements a DBI-compliant interface to MariaDB (\url{https://mariadb.org/}) and MySQL (\url{https://www.mysql.com/}) databases.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://rmariadb.r-dbi.org}
  \item \url{https://github.com/r-dbi/RMariaDB}
  \item \url{https://downloads.mariadb.org/connector-c/}
  \item Report bugs at \url{https://github.com/r-dbi/RMariaDB/issues}
}

}
\author{
\strong{Maintainer}: Kirill Müller \email{kirill@cynkra.com} (\href{https://orcid.org/0000-0002-1416-3412}{ORCID})

Authors:
\itemize{
  \item Jeroen Ooms (\href{https://orcid.org/0000-0002-4035-0289}{ORCID})
  \item David James
  \item Saikat DebRoy
  \item Hadley Wickham
  \item Jeffrey Horner
}

Other contributors:
\itemize{
  \item R Consortium [funder]
  \item RStudio [copyright holder]
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/connect.R, R/dbConnect_MariaDBDriver.R
\name{MariaDB}
\alias{MariaDB}
\alias{dbConnect_MariaDBDriver}
\alias{dbConnect,MariaDBDriver-method}
\title{Connect/disconnect to a MariaDB DBMS}
\usage{
MariaDB()

\S4method{dbConnect}{MariaDBDriver}(
  drv,
  dbname = NULL,
  username = NULL,
  password = NULL,
  host = NULL,
  unix.socket = NULL,
  port = 0,
  client.flag = 0,
  group = "rs-dbi",
  default.file = NULL,
  ssl.key = NULL,
  ssl.cert = NULL,
  ssl.ca = NULL,
  ssl.capath = NULL,
  ssl.cipher = NULL,
  ...,
  groups = NULL,
  load_data_local_infile = FALSE,
  bigint = c("integer64", "integer", "numeric", "character"),
  timeout = 10,
  timezone = "+00:00",
  timezone_out = NULL,
  reconnect = FALSE,
  mysql = NULL
)
}
\arguments{
\item{drv}{an object of class \linkS4class{MariaDBDriver} or
\linkS4class{MariaDBConnection}.}

\item{dbname}{string with the database name or NULL. If not NULL, the
connection sets the default database to this value.}

\item{username, password}{Username and password. If username omitted,
defaults to the current user. If password is omitted, only users
without a password can log in.}

\item{host}{string identifying the host machine running the MariaDB server or
NULL. If NULL or the string \code{"localhost"}, a connection to the local
host is assumed.}

\item{unix.socket}{(optional) string of the unix socket or named pipe.}

\item{port}{(optional) integer of the TCP/IP default port.}

\item{client.flag}{(optional) integer setting various MariaDB client flags,
see \link{Client-flags} for details.}

\item{group}{string identifying a section in the \code{default.file} to use
for setting authentication parameters (see \code{\link[=MariaDB]{MariaDB()}}).}

\item{default.file}{string of the filename with MariaDB client options,
only relevant if \code{groups} is given. The default value depends on the
operating system (see references), on Linux and OS X the files
\verb{~/.my.cnf} and \verb{~/.mylogin.cnf} are used. Expanded with \code{\link[=normalizePath]{normalizePath()}}.}

\item{ssl.key}{(optional) string of the filename of the SSL key file to use.
Expanded with \code{\link[=normalizePath]{normalizePath()}}.}

\item{ssl.cert}{(optional) string of the filename of the SSL certificate to
use. Expanded with \code{\link[=normalizePath]{normalizePath()}}.}

\item{ssl.ca}{(optional) string of the filename of an SSL certificate
authority file to use. Expanded with \code{\link[=normalizePath]{normalizePath()}}.}

\item{ssl.capath}{(optional) string of the path to a directory containing
the trusted SSL CA certificates in PEM format. Expanded with
\code{\link[=normalizePath]{normalizePath()}}.}

\item{ssl.cipher}{(optional) string list of permitted ciphers to use for SSL
encryption.}

\item{...}{Unused, needed for compatibility with generic.}

\item{groups}{deprecated, use \code{group} instead.}

\item{load_data_local_infile}{Set to \code{TRUE} to use \verb{LOAD DATA LOCAL INFILE}
in \code{\link[DBI:dbWriteTable]{DBI::dbWriteTable()}} and \code{\link[DBI:dbAppendTable]{DBI::dbAppendTable()}} by default.
This capability is disabled by default on the server side
for recent versions of MySQL Server.}

\item{bigint}{The R type that 64-bit integer types should be mapped to,
default is \link[bit64:bit64-package]{bit64::integer64}, which allows the full range of 64 bit
integers.}

\item{timeout}{Connection timeout, in seconds. Use \code{Inf} or a negative value
for no timeout.}

\item{timezone}{(optional) time zone for the connection,
the default corresponds to UTC.
Set this argument if your server or database is configured with a different
time zone than UTC.
Set to \code{NULL} to automatically determine the server time zone.}

\item{timezone_out}{The time zone returned to R.
The default is to use the value of the \code{timezone} argument,
\code{"+00:00"} is converted to \code{"UTC"}
If you want to display datetime values in the local timezone,
set to \code{\link[=Sys.timezone]{Sys.timezone()}} or \code{""}.
This setting does not change the time values returned, only their display.}

\item{reconnect}{(experimental) Set to \code{TRUE} to use \code{MYSQL_OPT_RECONNECT} to enable
automatic reconnection. This is experimental and could be dangerous if the connection
is lost in the middle of a transaction.}

\item{mysql}{Set to \code{TRUE}/\code{FALSE} to connect to a MySQL server or to a MariaDB server,
respectively.
The \pkg{RMariaDB} package supports both MariaDB and MySQL servers, but the SQL dialect
and other details vary.
The default is to assume MariaDB if the version is >= 10.0.0, and MySQL otherwise.}
}
\description{
These methods are straight-forward implementations of the corresponding
generic functions.
}
\section{Time zones}{

MySQL and MariaDB support named time zones,
they must be installed on the server.
See \url{https://dev.mysql.com/doc/mysql-g11n-excerpt/8.0/en/time-zone-support.html}
for more details.
Without installation, time zone support is restricted to UTC offset,
which cannot take into account DST offsets.
}

\section{Secure passwords}{

Avoid storing passwords hard-coded in the code, use e.g. the \pkg{keyring}
package to store and retrieve passwords in a secure way.

The MySQL client library (but not MariaDB) supports a \code{.mylogin.cnf} file
that can be passed in the \code{default.file} argument.
This file can contain an obfuscated password, which is not a secure way
to store passwords but may be acceptable if the user is aware of the
restrictions.
The availability of this feature depends on the client library used
for compiling the \pkg{RMariaDB} package.
Windows and macOS binaries on CRAN are compiled against the MariaDB Connector/C
client library which do not support this feature.
}

\examples{
if (mariadbHasDefault()) {
  # connect to a database and load some data
  con <- dbConnect(RMariaDB::MariaDB(), dbname = "test")
  dbWriteTable(con, "USArrests", datasets::USArrests, temporary = TRUE)

  # query
  rs <- dbSendQuery(con, "SELECT * FROM USArrests")
  d1 <- dbFetch(rs, n = 10)      # extract data in chunks of 10 rows
  dbHasCompleted(rs)
  d2 <- dbFetch(rs, n = -1)      # extract all remaining data
  dbHasCompleted(rs)
  dbClearResult(rs)
  dbListTables(con)

  # clean up
  dbDisconnect(con)
}
\dontrun{
# Connect to a MariaDB database running locally
con <- dbConnect(RMariaDB::MariaDB(), dbname = "mydb")
# Connect to a remote database with username and password
con <- dbConnect(RMariaDB::MariaDB(),
  host = "mydb.mycompany.com",
  user = "abc", password = "def"
)
# But instead of supplying the username and password in code, it's usually
# better to set up a group in your .my.cnf (usually located in your home
# directory). Then it's less likely you'll inadvertently share them.
con <- dbConnect(RMariaDB::MariaDB(), group = "test")

# To connect to a remote database and require the use of SSL
(and an example of using environment variables for your sensitive info)
con <- dbConnect(RMariaDB::MariaDB(),
  dbname = Sys.getenv('DB_NAME'),
  host = Sys.getenv('DB_HOST'),
  user = Sys.getenv('DB_USER'),
  password = Sys.getenv('DB_PASSWORD'),
  client.flag = CLIENT_SSL
)

# Always cleanup by disconnecting the database
dbDisconnect(con)
}

# All examples use the rs-dbi group by default.
if (mariadbHasDefault()) {
  con <- dbConnect(RMariaDB::MariaDB(), dbname = "test")
  con
  dbDisconnect(con)
}

}
\references{
Configuration files: https://mariadb.com/kb/en/library/configuring-mariadb-with-mycnf/
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MariaDBDriver.R, R/dbGetInfo_MariaDBDriver.R,
%   R/dbIsValid_MariaDBDriver.R, R/dbUnloadDriver_MariaDBDriver.R
\docType{class}
\name{MariaDBDriver-class}
\alias{MariaDBDriver-class}
\alias{dbGetInfo_MariaDBDriver}
\alias{dbGetInfo,MariaDBDriver-method}
\alias{dbIsValid_MariaDBDriver}
\alias{dbIsValid,MariaDBDriver-method}
\alias{dbUnloadDriver_MariaDBDriver}
\alias{dbUnloadDriver,MariaDBDriver-method}
\title{Class MariaDBDriver with constructor MariaDB.}
\usage{
\S4method{dbGetInfo}{MariaDBDriver}(dbObj, ...)

\S4method{dbIsValid}{MariaDBDriver}(dbObj, ...)

\S4method{dbUnloadDriver}{MariaDBDriver}(drv, ...)
}
\description{
An MariaDB driver implementing the R database (DBI) API.
This class should always be initialized with the \code{\link[=MariaDB]{MariaDB()}} function.
It returns a singleton that allows you to connect to MariaDB.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dbAppendTable_MariaDBConnection.R,
%   R/dbExistsTable_MariaDBConnection_character.R,
%   R/dbListObjects_MariaDBConnection_ANY.R, R/dbListTables_MariaDBConnection.R,
%   R/dbReadTable_MariaDBConnection_character.R,
%   R/dbRemoveTable_MariaDBConnection_character.R,
%   R/dbWriteTable_MariaDBConnection_character_character.R,
%   R/dbWriteTable_MariaDBConnection_character_data.frame.R, R/table.R
\name{mariadb-tables}
\alias{mariadb-tables}
\alias{dbAppendTable_MariaDBConnection}
\alias{dbAppendTable,MariaDBConnection-method}
\alias{dbExistsTable_MariaDBConnection_character}
\alias{dbExistsTable,MariaDBConnection,character-method}
\alias{dbListObjects_MariaDBConnection_ANY}
\alias{dbListObjects,MariaDBConnection-method}
\alias{dbListTables_MariaDBConnection}
\alias{dbListTables,MariaDBConnection-method}
\alias{dbReadTable_MariaDBConnection_character}
\alias{dbReadTable,MariaDBConnection,character-method}
\alias{dbRemoveTable_MariaDBConnection_character}
\alias{dbRemoveTable,MariaDBConnection,character-method}
\alias{dbWriteTable_MariaDBConnection_character_character}
\alias{dbWriteTable,MariaDBConnection,character,character-method}
\alias{dbWriteTable_MariaDBConnection_character_data.frame}
\alias{dbWriteTable,MariaDBConnection,character,data.frame-method}
\title{Read and write MariaDB tables.}
\usage{
\S4method{dbAppendTable}{MariaDBConnection}(conn, name, value, ..., row.names = NULL)

\S4method{dbExistsTable}{MariaDBConnection,character}(conn, name, ...)

\S4method{dbListObjects}{MariaDBConnection}(conn, prefix = NULL, ...)

\S4method{dbListTables}{MariaDBConnection}(conn, ...)

\S4method{dbReadTable}{MariaDBConnection,character}(conn, name, ..., row.names = FALSE, check.names = TRUE)

\S4method{dbRemoveTable}{MariaDBConnection,character}(conn, name, ..., temporary = FALSE, fail_if_missing = TRUE)

\S4method{dbWriteTable}{MariaDBConnection,character,character}(
  conn,
  name,
  value,
  field.types = NULL,
  overwrite = FALSE,
  append = FALSE,
  header = TRUE,
  row.names = FALSE,
  nrows = 50,
  sep = ",",
  eol = "\\n",
  skip = 0,
  quote = "\\"",
  temporary = FALSE,
  ...
)

\S4method{dbWriteTable}{MariaDBConnection,character,data.frame}(
  conn,
  name,
  value,
  field.types = NULL,
  row.names = FALSE,
  overwrite = FALSE,
  append = FALSE,
  ...,
  temporary = FALSE
)
}
\arguments{
\item{conn}{a \linkS4class{MariaDBConnection} object, produced by
\code{\link[DBI:dbConnect]{DBI::dbConnect()}}}

\item{name}{a character string specifying a table name.}

\item{value}{A data frame.}

\item{...}{Unused, needed for compatibility with generic.}

\item{row.names}{Either \code{TRUE}, \code{FALSE}, \code{NA} or a string.

If \code{TRUE}, always translate row names to a column called "row_names".
If \code{FALSE}, never translate row names. If \code{NA}, translate
rownames only if they're a character vector.

A string is equivalent to \code{TRUE}, but allows you to override the
default name.

For backward compatibility, \code{NULL} is equivalent to \code{FALSE}.}

\item{prefix}{A fully qualified path in the database's namespace, or \code{NULL}.
This argument will be processed with \code{\link[DBI:dbUnquoteIdentifier]{dbUnquoteIdentifier()}}.
If given the method will return all objects accessible through this prefix.}

\item{check.names}{If \code{TRUE}, the default, column names will be
converted to valid R identifiers.}

\item{temporary}{If \code{TRUE}, creates a temporary table that expires
when the connection is closed. For \code{dbRemoveTable()}, only temporary
tables are considered if this argument is set to \code{TRUE}.}

\item{fail_if_missing}{If \code{FALSE}, \code{dbRemoveTable()} succeeds if the
table doesn't exist.}

\item{field.types}{Optional, overrides default choices of field types,
derived from the classes of the columns in the data frame.}

\item{overwrite}{a logical specifying whether to overwrite an existing table
or not. Its default is \code{FALSE}.}

\item{append}{a logical specifying whether to append to an existing table
in the DBMS.  If appending, then the table (or temporary table)
must exist, otherwise an error is reported. Its default is \code{FALSE}.}

\item{header}{logical, does the input file have a header line? Default is the
same heuristic used by \code{read.table()}, i.e., \code{TRUE} if the first
line has one fewer column that the second line.}

\item{nrows}{number of lines to rows to import using \code{read.table} from
the input file to create the proper table definition. Default is 50.}

\item{sep}{field separator character}

\item{eol}{End-of-line separator}

\item{skip}{number of lines to skip before reading data in the input file.}

\item{quote}{the quote character used in the input file (defaults to
\verb{\\"}.)}
}
\value{
A data.frame in the case of \code{dbReadTable()}; otherwise a logical
indicating whether the operation was successful.
}
\description{
These methods read or write entire tables from a MariaDB database.
}
\details{
When using \code{load_data_local_infile = TRUE} in \code{\link[DBI:dbConnect]{DBI::dbConnect()}},
pass \code{safe = FALSE} to \code{dbAppendTable()} to avoid transactions.
Because \verb{LOAD DATA INFILE} is used internally, this means that
rows violating primary key constraints are now silently ignored.
}
\note{
The data.frame returned by \code{dbReadTable()} only has
primitive data, e.g., it does not coerce character data to factors.
Temporary tables are ignored for \code{dbExistsTable()} and \code{dbListTables()} due to
limitations of the underlying C API. For this reason, a prior existence check
is performed only before creating a regular persistent table; an attempt to
create a temporary table with an already existing name will fail with a
message from the database driver.
}
\examples{
if (mariadbHasDefault()) {
  con <- dbConnect(RMariaDB::MariaDB(), dbname = "test")

  # By default, row names are written in a column to row_names, and
  # automatically read back into the row.names()
  dbWriteTable(con, "mtcars", mtcars[1:5, ], temporary = TRUE)
  dbReadTable(con, "mtcars")
  dbReadTable(con, "mtcars", row.names = FALSE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MariaDBResult.R, R/dbIsValid_MariaDBResult.R
\docType{class}
\name{MariaDBResult-class}
\alias{MariaDBResult-class}
\alias{dbIsValid_MariaDBResult}
\alias{dbIsValid,MariaDBResult-method}
\title{Class MariaDBResult}
\usage{
\S4method{dbIsValid}{MariaDBResult}(dbObj, ...)
}
\description{
MariaDB's query results class.  This classes encapsulates the result of an SQL
query or statement.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dbBegin_MariaDBConnection.R,
%   R/dbCommit_MariaDBConnection.R, R/dbRollback_MariaDBConnection.R,
%   R/transaction.R
\name{transactions}
\alias{transactions}
\alias{dbBegin_MariaDBConnection}
\alias{dbBegin,MariaDBConnection-method}
\alias{dbCommit_MariaDBConnection}
\alias{dbCommit,MariaDBConnection-method}
\alias{dbRollback_MariaDBConnection}
\alias{dbRollback,MariaDBConnection-method}
\title{DBMS Transaction Management}
\usage{
\S4method{dbBegin}{MariaDBConnection}(conn, ...)

\S4method{dbCommit}{MariaDBConnection}(conn, ...)

\S4method{dbRollback}{MariaDBConnection}(conn, ...)
}
\arguments{
\item{conn}{a \linkS4class{MariaDBConnection} object, as produced by
\code{\link[DBI:dbConnect]{DBI::dbConnect()}}.}

\item{...}{Unused.}
}
\description{
Commits or roll backs the current transaction in an MariaDB connection.
Note that in MariaDB DDL statements (e.g. \verb{CREATE TABLE}) cannot
be rolled back.
}
\examples{
if (mariadbHasDefault()) {
  con <- dbConnect(RMariaDB::MariaDB(), dbname = "test")
  df <- data.frame(id = 1:5)

  dbWriteTable(con, "df", df, temporary = TRUE)
  dbBegin(con)
  dbExecute(con, "UPDATE df SET id = id * 10")
  dbGetQuery(con, "SELECT id FROM df")
  dbRollback(con)

  dbGetQuery(con, "SELECT id FROM df")

  dbDisconnect(con)
}
}
